import numpy as np
from scipy.spatial.distance import cosine
from sklearn.metrics.pairwise import cosine_distances

from tqdm import tqdm

from sciterra.vectorization.vectorizer import Vectorizer
from sciterra.vectorization import scibert, sbert, word2vec, bow, gpt2

astro_corpus_1 = "src/tests/data/corpora/example.txt"
model_path_1 = "src/tests/data/models/word2vec_model_example.model"

abstract_str = "We use cosmological hydrodynamic simulations with stellar feedback from the FIRE (Feedback In Realistic Environments) project to study the physical nature of Lyman limit systems (LLSs) at z ≤ 1. At these low redshifts, LLSs are closely associated with dense gas structures surrounding galaxies, such as galactic winds, dwarf satellites and cool inflows from the intergalactic medium. Our analysis is based on 14 zoom-in simulations covering the halo mass range M<SUB>h</SUB> ≈ 10<SUP>9</SUP>-10<SUP>13</SUP> M<SUB>⊙</SUB> at z = 0, which we convolve with the dark matter halo mass function to produce cosmological statistics. We find that the majority of cosmologically selected LLSs are associated with haloes in the mass range 10<SUP>10</SUP> ≲ M<SUB>h</SUB> ≲ 10<SUP>12</SUP> M<SUB>⊙</SUB>. The incidence and H I column density distribution of simulated absorbers with columns in the range 10^{16.2} ≤ N_{H I} ≤ 2× 10^{20} cm<SUP>-2</SUP> are consistent with observations. High-velocity outflows (with radial velocity exceeding the halo circular velocity by a factor of ≳ 2) tend to have higher metallicities ([X/H] ∼ -0.5) while very low metallicity ([X/H] &lt; -2) LLSs are typically associated with gas infalling from the intergalactic medium. However, most LLSs occupy an intermediate region in metallicity-radial velocity space, for which there is no clear trend between metallicity and radial kinematics. The overall simulated LLS metallicity distribution has a mean (standard deviation) [X/H] = -0.9 (0.4) and does not show significant evidence for bimodality, in contrast to recent observational studies, but consistent with LLSs arising from haloes with a broad range of masses and metallicities."  # 252 tokens

##############################################################################
# SciBERT
##############################################################################


class TestSciBERTVectorizer:
    vectorizer = scibert.SciBERTVectorizer()  # don't pass 'mps' for CI tests
    embedding_dim = scibert.EMBEDDING_DIM

    def test_single_vector(self):
        embedding = TestSciBERTVectorizer.vectorizer.embed_documents([abstract_str])[
            "embeddings"
        ]

        # Check embedding is of correct type, shape, and has no nans
        assert isinstance(embedding, np.ndarray)
        assert embedding.shape == (1, TestSciBERTVectorizer.embedding_dim)
        assert not np.isnan(embedding).any()

    def test_identity_of_embeddings(self):
        embeddings = TestSciBERTVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]
        # check identity
        assert np.all(embeddings[0] == embeddings[1])

    def test_diff_seq_len_batch(self):
        embeddings = TestSciBERTVectorizer.vectorizer.embed_documents(
            [
                abstract_str,
                abstract_str,
                abstract_str[:194],
                abstract_str[:194],
            ]
        )["embeddings"]

        # Check that first two and second two embeddings are the same
        assert np.array_equal(embeddings[0], embeddings[1])
        assert np.array_equal(embeddings[2], embeddings[3])

        # Check that second and third embeddings are not the same
        assert not np.array_equal(embeddings[1], embeddings[2])

    def test_single_cosine_pair(self):
        embeddings = TestSciBERTVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]

        # Check that the cosine sim of doc w/ itself is 1
        # n.b., see sklearn.metrics.pairwise.cosine_similarity
        sim = float(1 - cosine(embeddings[0], embeddings[1]))
        assert np.isclose(sim, 1.0)

    def test_basic_cosine_matrix(self):
        # like pair above, but pretending that we have more than 2 publications.

        num_pubs = 300
        # n.b., 1000 typically takes 83.75s with mps; Colab cuda takes just 29s
        # github actions just uses cpu, so maybe don't waste time stress testing

        embeddings = np.array(
            [
                TestSciBERTVectorizer.vectorizer.embed_documents(
                    [abstract_str] * num_pubs
                )["embeddings"].flatten()
            ]
        )
        cosine_matrix = cosine_distances(embeddings, embeddings)
        assert np.all(cosine_matrix == 0)


##############################################################################
# SBERT
##############################################################################


class TestSBERTVectorizer:
    vectorizer = sbert.SBERTVectorizer()  # don't pass 'mps' for CI tests
    embedding_dim = sbert.EMBEDDING_DIM

    def test_single_vector(self):
        embedding = TestSBERTVectorizer.vectorizer.embed_documents([abstract_str])[
            "embeddings"
        ]

        # Check embedding is of correct type, shape, and has no nans
        assert isinstance(embedding, np.ndarray)
        assert embedding.shape == (1, TestSBERTVectorizer.embedding_dim)
        assert not np.isnan(embedding).any()

    def test_identity_of_embeddings(self):
        embeddings = TestSBERTVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]
        # check identity
        assert np.all(embeddings[0] == embeddings[1])

    def test_diff_seq_len_batch(self):
        embeddings = TestSBERTVectorizer.vectorizer.embed_documents(
            [
                abstract_str,
                abstract_str,
                abstract_str[:194],
                abstract_str[:194],
            ]
        )["embeddings"]

        # Check that first two and second two embeddings are the same
        assert np.array_equal(embeddings[0], embeddings[1])
        assert np.array_equal(embeddings[2], embeddings[3])

        # Check that second and third embeddings are not the same
        assert not np.array_equal(embeddings[1], embeddings[2])

    def test_single_cosine_pair(self):
        embeddings = TestSBERTVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]

        # Check that the cosine sim of doc w/ itself is 1
        # n.b., see sklearn.metrics.pairwise.cosine_similarity
        sim = float(1 - cosine(embeddings[0], embeddings[1]))
        assert np.isclose(sim, 1.0)

    def test_basic_cosine_matrix(self):
        # like pair above, but pretending that we have more than 2 publications.

        num_pubs = 300
        # n.b., 1000 typically takes 83.75s with mps; Colab cuda takes just 29s
        # github actions just uses cpu, so maybe don't waste time stress testing

        embeddings = np.array(
            [
                TestSBERTVectorizer.vectorizer.embed_documents(
                    [abstract_str] * num_pubs
                )["embeddings"].flatten()
            ]
        )
        cosine_matrix = cosine_distances(embeddings, embeddings)
        assert np.all(cosine_matrix == 0)


##############################################################################
# GPT-2
##############################################################################


class TestGPT2Vectorizer:
    vectorizer = gpt2.GPT2Vectorizer()  # don't pass 'mps' for CI tests
    embedding_dim = gpt2.EMBEDDING_DIM

    def test_single_vector(self):
        embedding = TestGPT2Vectorizer.vectorizer.embed_documents([abstract_str])[
            "embeddings"
        ]

        # Check embedding is of correct type, shape, and has no nans
        assert isinstance(embedding, np.ndarray)
        assert embedding.shape == (1, TestGPT2Vectorizer.embedding_dim)
        assert not np.isnan(embedding).any()

    def test_identity_of_embeddings(self):
        embeddings = TestGPT2Vectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]
        # check identity
        assert np.all(embeddings[0] == embeddings[1])

    def test_diff_seq_len_batch(self):
        embeddings = TestGPT2Vectorizer.vectorizer.embed_documents(
            [
                abstract_str,
                abstract_str,
                abstract_str[:194],
                abstract_str[:194],
            ]
        )["embeddings"]

        # Check that first two and second two embeddings are the same
        assert np.array_equal(embeddings[0], embeddings[1])
        assert np.array_equal(embeddings[2], embeddings[3])

        # Check that second and third embeddings are not the same
        assert not np.array_equal(embeddings[1], embeddings[2])

    def test_single_cosine_pair(self):
        embeddings = TestGPT2Vectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]

        # Check that the cosine sim of doc w/ itself is 1
        # n.b., see sklearn.metrics.pairwise.cosine_similarity
        sim = float(1 - cosine(embeddings[0], embeddings[1]))
        assert np.isclose(sim, 1.0)

    def test_basic_cosine_matrix(self):
        # like pair above, but pretending that we have more than 2 publications.

        num_pubs = 300
        # n.b., 1000 typically takes 83.75s with mps; Colab cuda takes just 29s
        # github actions just uses cpu, so maybe don't waste time stress testing

        embeddings = np.array(
            [
                TestGPT2Vectorizer.vectorizer.embed_documents(
                    [abstract_str] * num_pubs
                )["embeddings"].flatten()
            ]
        )
        cosine_matrix = cosine_distances(embeddings, embeddings)
        assert np.all(cosine_matrix == 0)


##############################################################################
# Word2Vec
##############################################################################


class TestWord2VecVectorizer:
    corpus_path = astro_corpus_1
    model_path = model_path_1
    vectorizer = word2vec.Word2VecVectorizer(
        corpus_path=corpus_path,
        model_path=model_path,
    )
    embedding_dim = word2vec.EMBEDDING_DIM

    def test_single_vector(self):
        embedding = TestWord2VecVectorizer.vectorizer.embed_documents([abstract_str])[
            "embeddings"
        ]

        # Check embedding is of correct type, shape, and has no nans
        assert isinstance(embedding, np.ndarray)
        assert embedding.shape == (1, TestWord2VecVectorizer.embedding_dim)
        assert not np.isnan(embedding).any()

    def test_failed_embedding(self):
        result = TestWord2VecVectorizer.vectorizer.embed_documents(
            ["outofvocabularyitem"]
        )

        # Check embedding is of correct type, shape, and has no nans
        assert np.array_equal(result["fail_indices"], np.array([0]))
        assert np.array_equal(result["success_indices"], np.array([]))
        assert np.array_equal(result["embeddings"], np.array([]))

    def test_identity_of_embeddings(self):
        embeddings = TestWord2VecVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]
        # check identity
        assert np.all(embeddings[0] == embeddings[1])

    def test_diff_seq_len_batch(self):
        embeddings = TestWord2VecVectorizer.vectorizer.embed_documents(
            [
                abstract_str,
                abstract_str,
                abstract_str[:194],
                abstract_str[:194],
            ]
        )["embeddings"]

        # Check that first two and second two embeddings are the same
        assert np.array_equal(embeddings[0], embeddings[1])
        assert np.array_equal(embeddings[2], embeddings[3])

        # Check that second and third embeddings are not the same
        assert not np.array_equal(embeddings[1], embeddings[2])

    def test_single_cosine_pair(self):
        embeddings = TestWord2VecVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]

        # Check that the cosine sim of doc w/ itself is 1
        # n.b., see sklearn.metrics.pairwise.cosine_similarity
        sim = float(1 - cosine(embeddings[0], embeddings[1]))
        assert np.isclose(sim, 1.0)

    def test_basic_cosine_matrix(self):
        # like pair above, but pretending that we have more than 2 publications.

        num_pubs = 300

        embeddings = np.array(
            [
                TestWord2VecVectorizer.vectorizer.embed_documents(
                    [abstract_str] * num_pubs
                )["embeddings"].flatten()
            ]
        )
        cosine_matrix = cosine_distances(embeddings, embeddings)
        assert np.all(cosine_matrix == 0)


##############################################################################
# BoW
##############################################################################


class TestBOWVectorizer:
    corpus_path = astro_corpus_1
    model_path = model_path_1
    vectorizer = bow.BOWVectorizer(
        corpus_path=corpus_path,
        model_path=model_path,
    )
    embedding_dim = vectorizer.embedding_dim

    def test_single_vector(self):
        embedding = TestBOWVectorizer.vectorizer.embed_documents([abstract_str])[
            "embeddings"
        ]

        # Check embedding is of correct type, shape, and has no nans
        assert isinstance(embedding, np.ndarray)
        assert embedding.shape == (1, TestBOWVectorizer.embedding_dim)
        assert not np.isnan(embedding).any()

    def test_failed_embedding(self):
        result = TestBOWVectorizer.vectorizer.embed_documents(
            [abstract_str, "outofvocabularyitem"]
        )

        # Check embedding is of correct type, shape, and has no nans
        assert np.array_equal(result["fail_indices"], np.array([1]))
        assert np.array_equal(result["success_indices"], np.array([0]))
        assert result["embeddings"].shape == (1, TestBOWVectorizer.embedding_dim)

    def test_identity_of_embeddings(self):
        embeddings = TestWord2VecVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]
        # check identity
        assert np.all(embeddings[0] == embeddings[1])

    def test_diff_seq_len_batch(self):
        embeddings = TestBOWVectorizer.vectorizer.embed_documents(
            [
                abstract_str,
                abstract_str,
                abstract_str[:194],
                abstract_str[:194],
            ]
        )["embeddings"]

        # Check that first two and second two embeddings are the same
        assert np.array_equal(embeddings[0], embeddings[1])
        assert np.array_equal(embeddings[2], embeddings[3])

        # Check that second and third embeddings are not the same
        assert not np.array_equal(embeddings[1], embeddings[2])

    def test_single_cosine_pair(self):
        embeddings = TestWord2VecVectorizer.vectorizer.embed_documents(
            [abstract_str, abstract_str]
        )["embeddings"]

        # Check that the cosine sim of doc w/ itself is 1
        # n.b., see sklearn.metrics.pairwise.cosine_similarity
        sim = float(1 - cosine(embeddings[0], embeddings[1]))
        assert np.isclose(sim, 1.0)

    def test_basic_cosine_matrix(self):
        # like pair above, but pretending that we have more than 2 publications.

        num_pubs = 300

        embeddings = np.array(
            [
                TestWord2VecVectorizer.vectorizer.embed_documents(
                    [abstract_str] * num_pubs
                )["embeddings"].flatten()
            ]
        )
        cosine_matrix = cosine_distances(embeddings, embeddings)
        assert np.all(cosine_matrix == 0)



"""Test basic pipeline functionality with each librarian."""

from sciterra.librarians import adslibrarian, s2librarian


##############################################################################
# Semantic Scholar
##############################################################################


class TestSemanticScholarLibrarian:
    librarian = s2librarian.SemanticScholarLibrarian()

    def test_semanticscholar_convert_single(self):
        identifier = "DOI:10.1093/mnras/stx952"
        paper_id = "f2c251056dee4c6f9130b31e5e3e4b3296051c49"
        papers = TestSemanticScholarLibrarian.librarian.get_publications(
            [identifier],
            convert=False,
        )
        pub = TestSemanticScholarLibrarian.librarian.convert_publication(
            paper=papers[0]
        )
        assert papers[0].paperId == paper_id
        assert papers[0].abstract == pub.abstract
        assert papers[0].paperId == pub.identifier  # n.b., paperId != DOI id

    def test_semanticscholar_convert_parallel(self):
        identifier = "DOI:10.1093/mnras/stx952"
        papers = TestSemanticScholarLibrarian.librarian.get_publications(
            [identifier],
            convert=False,
        )

        converted_pubs = TestSemanticScholarLibrarian.librarian.convert_publications(
            papers
        )
        assert papers[0].abstract == converted_pubs[0].abstract

    def test_semanticscholar_convert_100(self):
        identifiers = ["DOI:10.1093/mnras/stx952"] * 100
        paper_id = "f2c251056dee4c6f9130b31e5e3e4b3296051c49"

        papers = TestSemanticScholarLibrarian.librarian.get_publications(
            identifiers,
            convert=False,
        )

        converted = TestSemanticScholarLibrarian.librarian.convert_publications(
            papers,
        )
        assert len(converted) == 100
        assert all([pub.identifier == paper_id for pub in converted])

    def test_semanticscholar_single_query(self):
        # construct an atlas w a single identifier
        identifier = "DOI:10.1093/mnras/stx952"
        paper_id = "f2c251056dee4c6f9130b31e5e3e4b3296051c49"

        pubs = TestSemanticScholarLibrarian.librarian.get_publications(
            [identifier],
        )
        assert len(pubs) == 1

        # assumes converted
        assert pubs[0].identifier == paper_id

    def test_semanticscholar_100_query(self):
        identifiers = ["DOI:10.1093/mnras/stx952"] * 100
        paper_id = "f2c251056dee4c6f9130b31e5e3e4b3296051c49"

        pubs = TestSemanticScholarLibrarian.librarian.get_publications(
            identifiers,
        )

        assert len(pubs) == 100
        assert all([pub.identifier == paper_id for pub in pubs])


##############################################################################
# ADS
##############################################################################


class TestADSLibrarian:
    librarian = adslibrarian.ADSLibrarian()

    def test_ads_convert_single(self):
        identifier = "doi:10.1093/mnras/stx952"
        bibcode = "2017MNRAS.469.2292H"
        articles = TestADSLibrarian.librarian.get_publications(
            [identifier],
            convert=False,
        )
        pub = TestADSLibrarian.librarian.convert_publication(
            article=articles[0],
        )
        assert articles[0].bibcode == bibcode
        assert articles[0].abstract == pub.abstract
        assert articles[0].bibcode == pub.identifier  # n.b., bibcode != doi

    def test_ads_convert_parallel(self):
        identifier = "doi:10.1093/mnras/stx952"
        articles = TestADSLibrarian.librarian.get_publications(
            [identifier],
            convert=False,
        )

        converted_pubs = TestADSLibrarian.librarian.convert_publications(
            articles,
        )

        assert articles[0].abstract == converted_pubs[0].abstract

    def test_ads_convert_100(self):
        identifiers = ["doi:10.1093/mnras/stx952"] * 100
        bibcode = "2017MNRAS.469.2292H"

        articles = TestADSLibrarian.librarian.get_publications(
            identifiers,
            convert=False,
        )
        converted = TestADSLibrarian.librarian.convert_publications(
            articles,
        )

        assert len(converted) == 100
        assert all([pub.identifier == bibcode for pub in converted])

    def test_ads_single_query(self):
        # construct an atlas w a single identifier
        identifier = "doi:10.1093/mnras/stx952"
        bibcode = "2017MNRAS.469.2292H"

        pubs = TestADSLibrarian.librarian.get_publications(
            [identifier],
        )
        assert len(pubs) == 1

        # assumes converted
        assert pubs[0].identifier == bibcode

    def test_ads_100_query(self):
        identifiers = ["doi:10.1093/mnras/stx952"] * 100
        bibcode = "2017MNRAS.469.2292H"

        pubs = TestADSLibrarian.librarian.get_publications(
            identifiers,
        )

        assert len(pubs) == 100
        assert all([pub.identifier == bibcode for pub in pubs])



from .librarian import Librarian
from .adslibrarian import ADSLibrarian
from .s2librarian import SemanticScholarLibrarian

librarians = {
    "S2": SemanticScholarLibrarian,
    "ADS": ADSLibrarian,
}

"""Why is there not an ArxivLibrarian? For now, we are restricting to APIs that allow us to traverse literature graphs, and arxiv does not have one. While there is a useful pip-installable package for querying the arxiv api for papers, https://pypi.org/project/arxiv/, the returned object does not have information on references and citations. However, it may still be possible to obtain a large sample of publications with abstracts and submission dates (though no citation counts), because the arxiv API's limit for a single query is 300,000 results.
"""



"""Convenience functionality for organized expansions of an Atlas."""

import numpy as np

from .atlas import Atlas
from .cartography import Cartographer
from ..librarians import librarians
from ..vectorization import vectorizers
from typing import Callable

##############################################################################
# Iterative expansion helper function
##############################################################################


def iterate_expand(
    atl: Atlas,
    crt: Cartographer,
    atlas_dir: str,
    target_size: int,
    max_failed_expansions: int = 2,
    convergence_func: Callable[[Atlas], bool] = lambda _: False,
    center: str = None,
    n_pubs_per_exp_max: int = 4000,
    call_size: int = None,
    n_sources_max: int = None,
    record_pubs_per_update: bool = True,
    **project_kwargs,
) -> Atlas:
    """Build out an Atlas of publications, i.e. search for similar publications. This is done by iterating a sequence of [expand, save, project, save, track, save]. The convergence criterion is:

    `converged = len(atl) >= target_size or failures >= max_failed_expansions or convergence_func(atl)`

    Args:
        atl: the Atlas to expand

        crt: the Cartographer to use

        atlas_dir: the directory where Atlas binaries will be saved/loaded from

        target_size: stop iterating when we reach this number of publications in the Atlas

        max_failed_expansions: stop iterating when we fail to add new publications after this many successive iterations. Default is 2.

        convergence_func: a function taking an Atlas and returns True if the expansion loop should stop. This serves as an additional disjunctive convergence criterion besides `target_size` and `max_failed_expansions` to exit the expansion loop, which will be called at the convergence check after each expansion. For example, a redundant function to choose would be `lambda atl: len(atl) >= target_size`. By default returns False.

        center: (if given) center the search on this publication, preferentially searching related publications.

        n_pubs_per_exp_max: maximum number of publications allowed in a single expansion.

        call_size: maximum number of papers to call API for in one query; if less than `len(paper_ids)`, chunking will be performed.

        n_sources_max: maximum number of publications (already in the atlas) to draw references and citations from.

        record_pubs_per_update: whether to track all the publications that exist in the resulting atlas to `self.pubs_per_update`. This should only be set to `True` when you need to later filter by degree of convergence of the atlas.

        project_kwargs: keyword args propagated to every `Cartographer.project` call during iterate_expand; see `Cartographer.filter_by_func`.

    Returns:
        atl: the expanded Atlas
    """
    converged = False
    print_progress = lambda atl: print(  # view incremental progress
        f"Atlas has {len(atl)} publications and {len(atl.projection) if atl.projection is not None else 'None'} embeddings."
    )

    # Expansion loop
    failures = 0
    # Count previous iterations from loaded atlas as part of total
    its = len(atl.history["pubs_per_update"]) if atl.history is not None else 0
    while not converged:
        its += 1
        len_prev = len(atl)

        print(f"\nExpansion {its}\n-------------------------------")

        # Retrieve up to n_pubs_max citations and references.
        atl = crt.expand(
            atl,
            center=center,
            n_pubs_max=n_pubs_per_exp_max,
            call_size=call_size,
            n_sources_max=n_sources_max,
            record_pubs_per_update=record_pubs_per_update,
        )
        print_progress(atl)
        atl.save(atlas_dir)

        # Obtain document embeddings for all new abstracts.
        atl = crt.project(
            atl,
            verbose=True,
            record_pubs_per_update=record_pubs_per_update,
            **project_kwargs,
        )
        print_progress(atl)
        atl.save(atlas_dir)

        atl = crt.track(atl, calculate_convergence=True)
        atl.save(atlas_dir)

        if len_prev == len(atl):
            failures += 0
        else:
            failures = 0

        converged = (
            len(atl) >= target_size
            or failures >= max_failed_expansions
            or convergence_func(atl)
        )

    print(f"Exiting loop.")
    atl = crt.track(atl, calculate_convergence=True)
    atl.save(atlas_dir)

    converged_dict = {
        "target_size": len(atl) >= target_size,
        "max_failed_expansions": failures >= max_failed_expansions,
        "convergence_func": convergence_func(atl),
    }
    print(
        f"Expansion loop exited with atlas size {len(atl)} after {its} iterations meeting criteria: {converged_dict}."
    )

    return atl


def search_converged_ids(
    atl: Atlas,
    num_pubs_added: int,
    kernel_size: int = 16,
) -> list[str]:
    """Get all publication ids such that they did not change neighborhood identity over the duration of the addition of the last `num_pubs_added` publications added to the atlas during previous `Cartographer.expand` calls.

    Args:
        atl: Atlas to search for converged publications

        num_pubs_added: the number of publications that we require to have been added to the Atlas in order to compute convergence, in order to compute the find the update index in the Atlas history to filter by `>= kernel_size`.

        kernel_size: the minimum required size of the neighborhood that we will require to not have changed, w.r.t. `cond_d`. Default is 16.

    Returns:
        converged_pub_ids: a list of Publication identifiers corresponding to publications that have converged acording to the criterion.
    """
    kernels = atl.history["kernel_size"]  # shape `(num_pubs, max_update)`
    max_update = kernels.shape[1]  # total number of updates to an Atlas
    if not max_update:  # if there has not been any updates
        return []

    # Define the update index to be the update such that greater than or equal to `num_pubs_added` have been added between this update and the final one.
    # This is clearly dependent on the size of the update; for example, if each update in the history added many publications, or if `num_pubs_added` is small, this update_ind might in practice be max_update-1.
    update_ind = None
    history = atl.history["pubs_per_update"][:-1]  # exclude the final update
    for reverse_idx, ids_at_update in enumerate(history[::-1]):
        num_diff_from_final = len(set(atl.ids) - set(ids_at_update))
        if num_diff_from_final >= num_pubs_added:
            update_ind = max_update - reverse_idx - 1
            break

    if update_ind is None:
        print(
            f"Failed to find an update index > 0 such that {num_pubs_added} were added between then and the final update ({max_update-1})."
        )
        update_ind = 0
    else:
        print(
            f"Between update {update_ind} and the final update ({max_update-1}) there were {num_diff_from_final} publications added to the Atlas."
        )

    # Get all publications that have not changed neighborhoods up to kernel_size for the last con_d updates
    converged_filter = kernels[:, update_ind] >= kernel_size
    ids = np.array(atl.projection.index_to_identifier)
    converged_pub_ids = ids[converged_filter]

    criterion = dict(num_pubs_added=num_pubs_added, kernel_size=kernel_size)

    print(
        f"Convergence criterion {criterion.items()} (=> index {update_ind} out of {max_update} total updates) yields {len(converged_pub_ids)} ids out of {len(atl)} total ids."
    )

    return converged_pub_ids


class AtlasTracer:
    """Convenience data structure for bookkeeping expansions of an Atlas that reduces boilerplate and ensures an aligned update history between the Atlas and Cartographer."""

    def __init__(
        self,
        atlas_dir: str,
        atlas_center_bibtex: str,
        librarian_name: str,
        vectorizer_name: str,
        librarian_kwargs: dict = dict(),
        vectorizer_kwargs: dict = dict(),
    ) -> None:
        """Convenience wrapper data structure for tracked expansions, by aligning the history of a Cartographer with an Atlas.

        Args:
            atlas_dir: absolute path of the directory to save atlas data in, propogated to `Atlas.load` and `Atlas.save`

            atlas_center_bibtex: absolute path of the .bib file containing a single entry, which is the core, central publication, and this entry contains an identifier recognizable by the librarian corresponding to `librarian_name`.

            librarian_name: a str name of a librarian, one of `librarians.librarians.keys()`, e.g. 'S2' or 'ADS'.

            vectorizer_name: a str name of a vectorizer, one of `vectorization.vectorizers.keys()`, e.g. 'BOW' or 'SciBERT'.

            librarian_kwargs: keyword args propogated to a Librarian initialization; if values are `None` they will be omitted

            vectorizer_kwargs: keyword args propogated to a Vectorizer initialization; if values are `None` they will be omitted
        """
        ######################################################################
        # Initialize cartography tools
        ######################################################################

        # Get librarian
        librarian = librarians[librarian_name]
        librarian_kwargs = {k: v for k, v in librarian_kwargs.items() if v is not None}

        # Get vectorizer
        vectorizer = vectorizers[vectorizer_name]
        # Get vectorizer kwargs if they are not null in config
        vectorizer_kwargs = {
            k: v for k, v in vectorizer_kwargs.items() if v is not None
        }

        self.cartographer = Cartographer(
            librarian=librarian(
                **librarian_kwargs,
            ),
            vectorizer=vectorizer(
                **vectorizer_kwargs,
            ),
        )

        ######################################################################
        # Initialize/Load Atlas
        ######################################################################
        self.atlas_dir = atlas_dir
        # Load
        atl = Atlas.load(self.atlas_dir)
        if len(atl):
            print(
                f"Loaded atlas has {len(atl)} publications and {len(atl.projection) if atl.projection is not None else 'None'} embeddings.\n"
            )
            # Crucial step: align the history of crt with atl
            if atl.history is not None:
                self.cartographer.pubs_per_update = atl.history["pubs_per_update"]
                print(
                    f"Loaded atlas at expansion iteration {len(atl.history['pubs_per_update'])}."
                )
        else:
            print(f"Initializing atlas.")

            # Get the bibtex file containing the seed publication
            bibtex_fp = atlas_center_bibtex

            # Get center from file
            atl_center = self.cartographer.bibtex_to_atlas(bibtex_fp)
            atl_center = self.cartographer.project(atl_center)

            num_entries = len(atl_center.publications.values())
            if num_entries > 1:
                raise Exception(
                    f"To build out a centered atlas, the center is specified by loading a bibtex file with a single entry. Found {num_entries} entries in {bibtex_fp}"
                )

            # Set the atlas center
            atl = atl_center
            (atl.center,) = atl.publications.keys()

        self.atlas = atl
        self.atlas.save(atlas_dirpath=self.atlas_dir)

    def expand_atlas(
        self,
        target_size: int,
        **kwargs,
    ) -> None:
        """Start or continue the expansion of the Atlas by calling `iterate_expand` with aligned Cartographer and Atlas, by default centered on atl.center.

        Args:
            target_size: stop iterating expansion when Atlas contains this many publications; argument propagated to `iterate_expand`.

            kwargs: keyword args propagated to `iterate_expand`.
        """

        if "center" not in kwargs:
            kwargs["center"] = self.atlas.center

        iterate_expand(
            self.atlas,
            self.cartographer,
            self.atlas_dir,
            target_size,
            **kwargs,
        )



"""We use a simple word2vec model that gets a document vector by averaging all words in the document.

Since we are getting vectors for scientific documents, we must load a vocabulary to train the model from scratch. Therefore we define different subclasses for each scientific field, which may differ substantially by vocabulary.

There exists a Doc2Vec module by gensim, but we're mainly interested in a simple baseline to compare with sophisticated embeddings.

Links:
    gensim: https://radimrehurek.com/gensim/auto_examples/tutorials/run_word2vec.html#
"""

import os
import time

import numpy as np

from .vectorizer import Vectorizer
from .preprocessing import CustomPreprocessor
from tqdm import tqdm
from typing import Callable

from gensim.models import Word2Vec

from multiprocessing import cpu_count


EMBEDDING_DIM = 300

# Training data for vocabulary

current_file_abs_path = os.path.dirname(os.path.abspath(__file__))
corpora_path = os.path.join(current_file_abs_path, "corpora")
ASTROPHYSICS_CORPUS = "astro_small.txt"
DEFAULT_CORPUS = os.path.join(corpora_path, ASTROPHYSICS_CORPUS)


class Word2VecVectorizer(Vectorizer):
    def __init__(
        self,
        corpus_path: str,
        model_path: str = None,
        vector_size: int = EMBEDDING_DIM,
        window: int = 5,
        min_count: int = 2,
        workers: int = cpu_count(),
        epochs: int = 10,
        sg: bool = True,  # use skipgram for large datasets
        tokenizer: Callable[[str], list[str]] = None,
        **kwargs,
    ) -> None:
        """Construct a Word2Vec based document embedding model from a corpus."""
        super().__init__()

        if tokenizer is None:
            preprocessor = CustomPreprocessor()
            self.tokenizer = preprocessor.custom_preprocess

        if (model_path is None) or (not os.path.exists(model_path)):
            start = time.time()
            # Assume the file is line-based, and one document per line
            print(
                f"Loading and tokenizing data from {corpus_path} for vocabulary and training..."
            )
            sentences = [
                self.tokenizer(line)
                for line in tqdm(open(corpus_path), desc="tokenizing lines")
            ]

            print(f"Training Word2Vec model...")
            model = Word2Vec(
                sentences=sentences,
                vector_size=vector_size,
                window=window,
                min_count=min_count,
                workers=workers,
                epochs=epochs,
                sg=sg,
            )
            duration = time.time() - start
            print(f"Loaded corpus and trained model in {duration:.2f} seconds.")
        else:
            print(f"Loading saved Word2Vec model from {model_path}.")
            model = Word2Vec.load(model_path)

        self.model = model

        # We don't plan to train the model any further, so we call `init_sims` to make the model much more memory-efficient
        # If `replace` is set, forget the original vectors and only keep the normalized ones = saves lots of memory!
        self.model.init_sims(replace=True)

        # write model to disk to load later and save time
        if model_path is not None:
            print(f"Saving Word2Vec model at {model_path}.")
            self.model.save(model_path)

    def embed_documents(self, docs: list[str], **kwargs) -> np.ndarray:
        """Embed a list of documents (raw text) into word2vec document vectors by averaging the word vectors in each of the documents.

        Since there's no speedup via batching like there is in pytorch models, we iterate one document at a time.
        """
        means = []
        success_indices = []
        failed_indices = []
        for i, doc in tqdm(
            enumerate(docs),
            desc="embedding documents",
            leave=True,
            total=len(docs),
        ):
            mean = np.mean(
                [
                    self.model.wv[word]
                    for word in self.tokenizer(doc)
                    if word in self.model.wv
                ],  # shape `(300,)`
                axis=0,
            )
            # OOV items become NaN
            if np.isnan(mean).any():
                failed_indices.append(i)
            else:
                means.append(mean)
                success_indices.append(i)

        return {
            "embeddings": np.array(means),
            "success_indices": np.array(success_indices, dtype=int),
            "fail_indices": np.array(failed_indices, dtype=int),
        }



from .gpt2 import GPT2Vectorizer
from .scibert import SciBERTVectorizer
from .sbert import SBERTVectorizer
from .word2vec import Word2VecVectorizer
from .bow import BOWVectorizer

vectorizers = {
    "GPT2": GPT2Vectorizer,
    "SciBERT": SciBERTVectorizer,
    "SBERT": SBERTVectorizer,
    "Word2Vec": Word2VecVectorizer,
    "BOW": BOWVectorizer,
}



"""Main container object for a large library of publications.
"""

import os
import warnings
import numpy as np
import pandas as pd

from typing import Any

from .publication import Publication
from ..vectorization.projection import Projection
from ..misc.utils import read_pickle, write_pickle, get_verbose, custom_formatwarning

warnings.formatwarning = custom_formatwarning


class Atlas:

    """Data structure for storing publications.

    `self.projection`: the Projection object containing the embeddings of all publications and their mapping to str identifiers.

    `self.bad_ids`: a list of identifiers that have failed for some reason or other during an expansion, and will be excluded from subsequent expansions.

    `self.history`: dict of the form {'pubs_per_update': list[list[str]], 'kernel_size': np.ndarray of ints of shape `(num_pubs, last_update)` where last_update <= the total number of expansions performed.}

    `self.center`: the core, central Publication identifier repeatedly passed to `cartography.Cartographer.expand`. Default is None, which means the Atlas has no internal record of the central publication.
    """

    def __init__(
        self,
        publications: list[Publication],
        projection: Projection = None,
        bad_ids: set[str] = set(),
        history: dict[str, Any] = None,
        center: str = None,
    ) -> None:
        if not isinstance(publications, list):
            raise ValueError

        self.publications: dict[str, Publication] = {
            str(pub): pub for pub in publications
        }
        self.projection = projection
        self.bad_ids = bad_ids
        self.history = history
        self.center = center

    ######################################################################
    # Lookup    ######################################################################

    def __getitem__(self, identifier: str) -> Publication:
        """Get a publication given its identifier.

        Raises:
            ValueError: the identifier is not in the Atlas.
        """
        if identifier in self.publications:
            return self.publications[identifier]
        raise ValueError(f"Identifier {identifier} not in Atlas.")

    @property
    def ids(self) -> list[str]:
        """Get a list of all the publication identifiers in the Atlas."""
        return list(self.publications.keys())

    ######################################################################
    # File I/O
    ######################################################################

    def save(
        self,
        atlas_dirpath: str,
        overwrite: bool = True,
    ) -> None:
        """Write the Atlas to a directory containing a .pkl binary for each attribute.

        Warnings cannot be silenced.

        Args:
            atlas_dirpath: path of directory to save files to.
        """

        # Create directory as needed, or overwrite existing files
        if os.path.isdir(atlas_dirpath):
            if not overwrite:
                warnings.warn(
                    f"Skipping overwrite of atlas data found at {atlas_dirpath}."
                )
                return
        else:
            warnings.warn(
                f"Recursively creating atlas data directory at {atlas_dirpath}."
            )
            os.makedirs(atlas_dirpath)

        attributes = {
            k: getattr(self, k)
            for k in [
                "publications",
                "projection",
                "bad_ids",
                "history",
                "center",
            ]
        }

        for attribute in attributes:
            if getattr(self, attribute) is not None:
                # write the list version to be consistent with load and constructor
                if attribute == "publications":
                    attributes[attribute] = list(self.publications.values())

                fn = f"{attribute}.pkl"
                fp = os.path.join(atlas_dirpath, fn)
                if os.path.isfile(fp):
                    warnings.warn(f"Overwriting existing file at {fp}.")
                else:
                    warnings.warn(f"Writing to {fp}.")
                write_pickle(fp, attributes[attribute])
            else:
                warnings.warn(f"No {attribute} to save, skipping.")

    @classmethod
    def load(
        cls,
        atlas_dirpath: str,
    ):
        """Load an Atlas object from a directory containing the .pkl binary for each attribute.

        Warnings cannot be silenced.

        Args:
            atlas_dirpath: directory where .pkl binaries will be read from

        """
        attributes = {
            k: None
            for k in [
                "publications",
                "projection",
                "bad_ids",
                "history",
                "center",
            ]
        }
        for attribute in attributes:
            fn = f"{attribute}.pkl"
            fp = os.path.join(atlas_dirpath, fn)
            if os.path.isfile(fp):
                attributes[attribute] = read_pickle(fp)
            else:
                warnings.warn(f"No {attribute} to read, skipping.")

        if attributes["publications"] is None:
            warnings.warn("Loading empty atlas.")
            attributes["publications"] = list()

        return cls(**{k: v for k, v in attributes.items() if v is not None})

    ######################################################################
    # Other
    ######################################################################

    def __len__(self) -> int:
        """Get length of the Atlas."""
        return len(self.publications)

    def __eq__(self, __value: object) -> bool:
        return (
            self.publications == __value.publications
            and self.projection == __value.projection
        )



"""Bag of words document embedder. Unlike cc vectorization, we fix the dimension of the embeddings to be the same; this requires us to fix the vocabulary, so for consistency we do so via the same method as the Word2Vec vocabulary construction."""


import os
import time

import numpy as np
from numpy import ndarray

from .vectorizer import Vectorizer
from .word2vec import Word2VecVectorizer
from tqdm import tqdm

from sklearn.feature_extraction.text import CountVectorizer


from multiprocessing import cpu_count


# Training data for vocabulary

current_file_abs_path = os.path.dirname(os.path.abspath(__file__))
corpora_path = os.path.join(current_file_abs_path, "corpora")
ASTROPHYSICS_CORPUS = "astro_small.txt"
DEFAULT_CORPUS = os.path.join(corpora_path, ASTROPHYSICS_CORPUS)


class BOWVectorizer(Vectorizer):
    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Construct a bag-of-words document vectorizer."""
        # this part is slow because involves training the word2vec model
        self.word2vec_vectorizer = Word2VecVectorizer(*args, **kwargs)
        self.vocabulary = self.word2vec_vectorizer.model.wv.key_to_index
        self.embedding_dim = len(self.vocabulary)
        self.count_vectorizer = CountVectorizer(vocabulary=self.vocabulary)

    def embed_documents(self, docs: list[str], **kwargs) -> dict[str, ndarray]:
        """Embed a list of documents (raw text) into bow document vectors using scikit-learn's CountVectorizer.

        Args:
            docs: the documents to embed.

        Returns:
            a numpy array of shape `(num_documents, len(self.vocabulary))`
        """
        # We could process all documents at once, but we sacrifice some speed to see a progress bar.
        embeddings = np.array(
            [
                self.count_vectorizer.transform([doc]).toarray().sum(axis=0)
                for doc in tqdm(docs, desc="embedding documents", leave=True)
            ]
        )

        # Find the indices of documents that had at least 1 in-vocab token
        success_indices = np.where(np.sum(embeddings, axis=1) > 0)[0]
        fail_indices = np.setdiff1d(np.arange(len(docs)), success_indices)

        return {
            "embeddings": embeddings[success_indices],
            "success_indices": success_indices,
            "fail_indices": fail_indices,
        }



import ads

from ads.search import Article
from datetime import date, datetime

from sciterra.mapping.publication import Publication
from ..mapping.publication import Publication
from .librarian import Librarian

from ..misc.utils import chunk_ids, keep_trying, get_verbose

from tqdm import tqdm

import warnings

CALL_SIZE = (
    50  # handles more than 2000, much better than S2; but easy to hit TooManyRequests
)
NUM_ATTEMPTS_PER_QUERY = 10

QUERY_FIELDS = [
    "bibcode",  # str
    "abstract",  # str
    "title",  # list
    "entry_date",  # datetime (earliest possible)
    "pubdate",  # a datetime
    "year",  # int
    "citation_count",
    "citation",  # list
    "reference",  # list
    "identifier",  # list of external ids
    "arxiv_class",  # list of arxiv classifiers; interestingly, returned even for DOIs that aren't arxiv
]

ALLOWED_EXCEPTIONS = (ads.exceptions.APIResponseError,)

EXTERNAL_IDS = [
    "DOI",  # returns a list
    "arXiv",  # returns a str
    "bibcode",  # returns a str, preferred
]


class ADSLibrarian(Librarian):
    def __init__(self) -> None:
        super().__init__()

    def bibtex_entry_identifier(self, bibtex_entry: dict) -> str:
        """Parse a bibtex entry for a usable identifier for querying ADS (see EXTERNAL_IDS)."""
        identifier = None
        if "bibcode" in bibtex_entry:
            identifier = bibtex_entry["bibcode"]
        elif "doi" in bibtex_entry:
            identifier = f"doi:{bibtex_entry['doi']}"
        elif "arxiv" in bibtex_entry:
            identifier = f"arxiv:{bibtex_entry['arxiv']}"
        return identifier

    def get_publications(
        self,
        bibcodes: list[str],
        *args,
        call_size: int = CALL_SIZE,
        n_attempts_per_query: int = NUM_ATTEMPTS_PER_QUERY,
        convert: bool = True,
        **kwargs,
    ) -> list[Publication]:
        """Use the NASA ADS python package, which calls the ADS API to retrieve publications.

        Args:
            bibcodes: the str ids required for querying. While it is possible to use one of EXTERNAL_IDS to query, if ADS returns a paper at all, it will return a bibcode, so it is preferred to use bibcodes.

            n_attempts_per_query: Number of attempts to access the API per query. Useful when experiencing connection issues.

            call_size: maximum number of papers to call API for in one query; if less than `len(bibcodes)`, chunking will be performed.

            convert: whether to convert each resulting ADS Article to sciterra Publications (True by default).

        Returns:
            the list of publications (or Papers)
        """
        bibcodes = list(bibcodes)

        if not bibcodes:
            return []

        if call_size is None:
            call_size = CALL_SIZE

        total = len(bibcodes)
        chunked_ids = chunk_ids(
            bibcodes,
            call_size=call_size,
        )

        if None in bibcodes:
            # any Nones should have been handled by this point
            raise Exception("Passed `bibcodes` contains None.")

        print(f"Querying ADS for {len(bibcodes)} total papers.")
        papers = []
        pbar = tqdm(desc=f"progress using call_size={call_size}", total=total)
        for ids in chunked_ids:

            @keep_trying(
                n_attempts=n_attempts_per_query,
                allowed_exceptions=ALLOWED_EXCEPTIONS,
                sleep_after_attempt=2,
            )
            def get_papers() -> list[Article]:
                return [
                    list(
                        ads.SearchQuery(
                            query_dict={
                                "q": query,
                                "fl": QUERY_FIELDS,
                            }
                        )
                    )[
                        0  # screw black, this is ugly
                    ]  # retrieve from generator
                    for query in ids
                ]

            papers.extend(get_papers())
            pbar.update(len(ids))

        pbar.close()

        if not convert:
            return papers
        return self.convert_publications(
            papers,
            *args,
            **kwargs,
        )

    def convert_publication(self, article: Article, *args, **kwargs) -> Publication:
        """Convert a ADS Article object to a sciterra.publication.Publication."""
        if article is None:
            return

        verbose = get_verbose(kwargs)

        # to be consistent with identifiers (e.g., to avoid storing the same publication twice), we always use the bibcode.
        identifier = article.bibcode

        def process_date(date_str: str) -> str:
            # sometimes there is extra data
            date_str = date_str[:10]  # e.g. yyyy-mm-dd
            # sometimes ads has 00 for month or day
            if date_str[-2:] == "00":
                date_str[-2:] = "01"
            if date_str[-5:-3] == "00":
                date_str[-5:-3] = "01"
            date_ = datetime.strptime(date_str, "%Y-%m-%d")
            return date_

        # Parse date from datetime or year
        if hasattr(article, "entry_date"):
            publication_date = process_date(article.entry_date)
        elif hasattr(article, "pubdate"):
            publication_date = process_date(article.pubdate)
        elif hasattr(article, "year"):
            publication_date = date(article.year, 1, 1)
        else:
            publication_date = None

        # get doi from Article identifiers
        # warning: ADS tracks two DOIs: for official and arxiv
        doi = None
        if hasattr(article, "identifier"):
            dois = [item for item in article.identifier if item[:3] == "10."]
            if dois:
                doi = dois[0]

        # Process citation data
        citations = article.citation
        references = article.reference

        citation_count = article.citation_count
        if (
            (citation_count is not None)
            and (citations is not None)
            and (citation_count != len(citations))
            and verbose
        ):
            warnings.warn(
                f"The length of the citations list ({len(citations)}) is different from citation_count ({citation_count})"
            )
            if "infer_citation_count" in kwargs and kwargs["infer_citation_count"]:
                if verbose:
                    warnings.warn("Setting citation_count = {len(citations)}.")
                citation_count = len(citations)

        # N.B.: for fields_of_study, manually annotate, and add arxiv classes.
        fields_of_study = [
            "ads_dummy_field"
        ]  # n.b. there is actually some diversity of fields beyond physics in ADS, e.g. theoretical biology.
        arxiv_classes = article.arxiv_class if article.arxiv_class is not None else []
        fields_of_study = fields_of_study + arxiv_classes
        fields_of_study = list(set(fields_of_study))

        data = {
            # primary fields
            "identifier": identifier,
            "abstract": article.abstract,
            "publication_date": publication_date,
            "citations": citations,
            "references": references,
            "citation_count": citation_count,
            "fields_of_study": fields_of_study,
            # additional fields
            "doi": doi,
            "title": article.title,
        }
        data = {k: v for k, v in data.items() if v is not None}

        return Publication(data)


# TODO: consider implementing `random_publications`
# We are interested in determining whether the distribution of density is normal, even when not doing the iterative similarity-based expansion loop. To do this, we need a sample of publications that aren't necessarily similar to each other.

# This has already been implemented by Zach in his `background_distribution` step, using cc and ADS.



import numpy as np


class Projection:
    """Basic wrapper for document embeddings and helper methods."""

    def __init__(
        self,
        identifier_to_index: dict[str, int],
        index_to_identifier: tuple[str],
        embeddings: np.ndarray,
    ) -> None:
        """Construct a Projection object, a bidirectional mapping from identifiers to document embeddings.

        Args:
            identifier_to_index: a dict mapping Publication identifiers to indices in the embedding matrix.

            index_to_identifier: a tuple mapping embedding indices to Publication identifiers.

            embeddings: ndarray of document embeddings of shape `(num_pubs, embedding_dim)`
        """
        self.identifier_to_index = identifier_to_index
        self.index_to_identifier = index_to_identifier
        self.embeddings = embeddings

    def indices_to_identifiers(self, indices) -> list[str]:
        """Retrieve the identifiers for a list of embedding matrix indices."""
        return [self.index_to_identifier[index] for index in indices]

    def identifiers_to_embeddings(self, identifiers: list[str]) -> np.ndarray:
        """Retrieve the document embeddings for a list of identifiers."""
        return self.embeddings[self.identifiers_to_indices(identifiers)]

    def identifiers_to_indices(self, identifiers: list[str]) -> np.ndarray:
        """Retrieve the embedding indices for a list of identifiers."""
        return np.array(
            [self.identifier_to_index[identifier] for identifier in identifiers]
        )

    def __len__(self) -> int:
        return len(self.identifier_to_index)

    def __eq__(self, __value: object) -> bool:
        return (
            np.array_equal(self.embeddings, __value.embeddings)
            and self.identifier_to_index == __value.identifier_to_index
            and self.index_to_identifier == __value.index_to_identifier
        )


######################################################################
# Merge projections
######################################################################


def merge(proj_a: Projection, proj_b: Projection) -> Projection:
    """Return the result of merging projection `proj_a` with projection `proj_b`.

    This adds to proj_a all embedding data contained in proj_b that is missing from proj_a. This means that the resulting projection can only be greater or equal in size to proj_a.
    """
    if proj_b is None or not len(proj_b):
        return proj_a

    # Get the data in the new projection missing from the old
    indices_missing = []
    identifiers_missing = []
    for id, idx in proj_b.identifier_to_index.items():
        if proj_a is None or id not in proj_a.identifier_to_index:
            indices_missing.append(idx)
            identifiers_missing.append(id)

    # Get just the missing embeddings
    embeddings_missing = np.array(
        [
            embedding
            for idx, embedding in enumerate(proj_b.embeddings)
            if idx in set(indices_missing)
        ]
    )

    # Concatenate index mapping and embeddings
    idx_to_ids_new = identifiers_missing
    embeddings_new = embeddings_missing
    if proj_a is not None:
        idx_to_ids_new = list(proj_a.index_to_identifier) + idx_to_ids_new
        embeddings_new = np.concatenate([proj_a.embeddings, embeddings_new])

    # Return a new projection
    return Projection(
        identifier_to_index={id: idx for idx, id in enumerate(idx_to_ids_new)},
        index_to_identifier=tuple(idx_to_ids_new),
        embeddings=embeddings_new,
    )


def get_empty_projection() -> Projection:
    """Construct a Projection with no data (but is not None)."""
    return Projection(
        identifier_to_index={},
        index_to_identifier=(),
        embeddings=np.array([[]]),  # 2D
    )



import warnings

from datetime import date

from typing import Any
from tqdm import tqdm


from sciterra.mapping.publication import Publication

from ..mapping.publication import Publication
from .librarian import Librarian
from ..misc.utils import chunk_ids, keep_trying, get_verbose

from semanticscholar import SemanticScholar
from semanticscholar.Paper import Paper

from requests.exceptions import ReadTimeout, ConnectionError
from semanticscholar.SemanticScholarException import ObjectNotFoundException

##############################################################################
# Constants
##############################################################################

# NOTE: semantic scholar will truncate total number of references, citations each at 10,000 for the entire batch.
QUERY_FIELDS = [
    "year",
    "abstract",
    "title",  # useful for inspection
    "externalIds",  # supports ArXiv, MAG, ACL, PubMed, Medline, PubMedCentral, DBLP, DOI
    "citationCount",
    "fieldsOfStudy",  # useful for scoping to a particular field
    "s2FieldsOfStudy",  # if above (annotated) is none we can still extract predicted field
    "url",  # as a possible external id
    "citations.externalIds",
    "citations.url",
    "references.externalIds",
    "references.url",
    "citationStyles",  # supports a very basic bibtex that we will augment
    "publicationDate",  # if available, type datetime.datetime (YYYY-MM-DD)
]

# The following types of IDs are supported
EXTERNAL_IDS = [
    "DOI",
    "ArXiv",
    "CorpusId",
    "MAG",
    "ACL",
    "PubMed",
    "Medline",
    "PubMedCentral",
    "DBLP",
    "URL",
]

ALLOWED_EXCEPTIONS = (
    Exception,  # "Internal Service Error"
    ReadTimeout,
    ConnectionError,
    ObjectNotFoundException,
)
CALL_SIZE = 10
NUM_ATTEMPTS_PER_QUERY = 50

##############################################################################
# Main librarian class
##############################################################################


class SemanticScholarLibrarian(Librarian):
    def __init__(
        self,
        api_key: str = None,
        api_key_fn: str = None,
    ) -> None:
        if api_key_fn is not None:
            print(f"Reading private api key from {api_key_fn}.")
            # Parse api_key_fn for 40-ch private key
            with open(api_key_fn, "r") as f:
                api_key = f.read()

        self.sch = SemanticScholar(api_key=api_key)
        super().__init__()

    def bibtex_entry_identifier(self, bibtex_entry: dict) -> str:
        """Parse a bibtex entry for a usable identifier for querying SemanticScholar (see EXTERNAL_IDS)."""
        identifier = None
        if "paper_id" in bibtex_entry:
            identifier = bibtex_entry["paper_id"]
        elif "corpus_id" in bibtex_entry:
            identifier = f"CorpusID:{bibtex_entry['corpus_id']}"
        elif "doi" in bibtex_entry:
            identifier = f"DOI:{bibtex_entry['doi']}"
        return identifier

    def get_publications(
        self,
        paper_ids: list[str],
        *args,
        call_size: int = CALL_SIZE,
        n_attempts_per_query: int = NUM_ATTEMPTS_PER_QUERY,
        convert: bool = True,
        **kwargs,
    ) -> list[Publication]:
        """Use the (unofficial) S2 python package, which calls the Semantic Scholar API to retrieve publications from the S2AG.

        Args:
            paper_ids: the str ids required for querying. While it is possible to use one of EXTERNAL_IDS to query, if SemanticScholar returns a paper at all, it will return a paperId, so it is preferred to use paperIds.

            n_attempts_per_query: Number of attempts to access the API per query. Useful when experiencing connection issues.

            call_size: maximum number of papers to call API for in one query; if less than `len(paper_ids)`, chunking will be performed. Maximum that S2 allows is 500.

            convert: whether to convert each resulting SemanticScholar Paper to sciterra Publications (True by default).

        Returns:
            the list of publications (or Papers)
        """
        paper_ids = list(paper_ids)

        if not paper_ids:
            return []

        if call_size is None:
            call_size = CALL_SIZE

        total = len(paper_ids)
        chunked_ids = chunk_ids(
            paper_ids,
            call_size=call_size,
        )

        if None in paper_ids:
            # any Nones should have been handled by this point
            raise Exception("Passed `paper_ids` contains None.")

        print(f"Querying Semantic Scholar for {len(paper_ids)} total papers.")
        papers = []
        pbar = tqdm(desc=f"progress using call_size={call_size}", total=total)
        for ids in chunked_ids:

            @keep_trying(
                n_attempts=n_attempts_per_query,
                allowed_exceptions=ALLOWED_EXCEPTIONS,
                sleep_after_attempt=2,
            )
            def get_papers() -> list[Paper]:
                if call_size > 1:
                    result = self.get_papers(
                        paper_ids=ids,
                        fields=QUERY_FIELDS,
                    )
                else:
                    # typically completes about 100 queries per minute.
                    result = [
                        self.get_paper(
                            paper_id=paper_id,
                            fields=QUERY_FIELDS,
                        )
                        for paper_id in ids
                    ]
                return result

            papers.extend(get_papers())
            pbar.update(len(ids))
        pbar.close()

        if not convert:
            return papers
        return self.convert_publications(  # may contain Nones!
            papers,
            *args,
            **kwargs,
        )

    def convert_publication(self, paper: Paper, *args, **kwargs) -> Publication:
        """Convert a SemanticScholar Paper object to a sciterra.publication.Publication."""
        if paper is None:
            return

        verbose = get_verbose(kwargs)

        # to be consistent with identifiers (e.g., to avoid storing the same publication twice), we always use the paperId.
        identifier = paper.paperId

        # Parse date from datetime or year
        if paper.publicationDate is not None:
            publication_date = paper.publicationDate.date()
        elif paper.year is not None:
            publication_date = date(paper.year, 1, 1)
        else:
            publication_date = None

        # Parse citations
        citations = None
        if paper.citations is not None:
            # convert citations/references from lists of Papers to identifiers
            citations = [
                paper.paperId for paper in paper.citations if paper.paperId is not None
            ]  # no point using recursion assuming identifier=paperId

        references = [
            paper.paperId for paper in paper.references if paper.paperId is not None
        ]

        # TODO: same with citationCount
        citation_count = paper.citationCount
        if citation_count != len(citations) and verbose:
            warnings.warn(
                f"The length of the citations list ({len(citations)}) is different from citation_count ({citation_count})"
            )

        # TODO: What if citations = []?
        if (
            "infer_citation_count" in kwargs
            and kwargs["infer_citation_count"]
            and verbose
        ):
            warnings.warn("Setting citation_count = {len(citations)}.")
            citation_count = len(citations)

        # Clobber together a field of study from the validated annoatation or s2's model-predicted fields
        primary_fields = (
            paper.fieldsOfStudy
            if hasattr(paper, "fieldsOfStudy") and paper.fieldsOfStudy is not None
            else []
        )
        addl_fields = (
            [entry["category"] for entry in paper.s2FieldsOfStudy]
            if hasattr(paper, "s2FieldsOfStudy") and paper.s2FieldsOfStudy is not None
            else []
        )
        fields_of_study = primary_fields + addl_fields
        fields_of_study = list(set(fields_of_study))
        if not fields_of_study:
            fields_of_study = None

        data = {
            # primary fields
            "identifier": identifier,
            "abstract": paper.abstract,
            "publication_date": publication_date,
            "citations": citations,
            "references": references,
            "citation_count": citation_count,
            "fields_of_study": fields_of_study,
            # additional fields
            "doi": paper.externalIds["DOI"] if "DOI" in paper.externalIds else None,
            "url": paper.url if hasattr(paper, "url") else None,
            "title": paper.title if hasattr(paper, "title") else None,
            "issn": paper.issn if hasattr(paper, "issn") else None,
        }
        data = {k: v for k, v in data.items() if v is not None}

        return Publication(data)

    # We write this minimally different function from SemanticScholar.get_papers so that others dont need to fork our version of semantic scholar.
    def get_papers(self, paper_ids: list[str], fields: list[str]):
        """Custom function for calling the S2 API that doesn't fail on empty results."""
        if not fields:
            fields = Paper.SEARCH_FIELDS

        base_url = self.sch.api_url + self.sch.BASE_PATH_GRAPH
        url = f"{base_url}/paper/batch"

        fields = ",".join(fields)
        parameters = f"&fields={fields}"

        payload = {"ids": paper_ids}

        data = self.sch._requester.get_data(
            url, parameters, self.sch.auth_header, payload
        )
        papers = [
            Paper(item) if item is not None else None for item in data
        ]  # added condition

        return papers

    # "
    def get_paper(self, paper_id: str, fields: list[str]):
        """Custom function for calling the S2 API that doesn't fail on empty results."""
        if not fields:
            fields = Paper.FIELDS

        base_url = self.sch.api_url + self.sch.BASE_PATH_GRAPH
        url = f"{base_url}/paper/{paper_id}"

        fields = ",".join(fields)
        parameters = f"&fields={fields}"

        data = self.sch._requester.get_data(url, parameters, self.sch.auth_header)
        paper = Paper(data) if data is not None else None  # added condition

        return paper



import numpy as np

from sciterra.mapping.cartography import Cartographer
from sciterra.mapping.tracing import iterate_expand, AtlasTracer, search_converged_ids
from sciterra.librarians.s2librarian import SemanticScholarLibrarian
from sciterra.vectorization import SciBERTVectorizer

from .test_cartography import (
    single_pub_bibtex_fp,
    atlas_dir,
    ten_pub_bibtex_fp,
)
from .test_vectorization import astro_corpus_1, model_path_1


class TestExpansion:
    def test_iterate_expand(self, tmp_path):
        librarian = SemanticScholarLibrarian()
        vectorizer = SciBERTVectorizer()
        crt = Cartographer(librarian, vectorizer)

        # Load single file from bibtex
        bibtex_fp = single_pub_bibtex_fp

        path = tmp_path / atlas_dir
        path.mkdir()

        # Construct Atlas
        atl = crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        center = pub.identifier

        iterate_expand(
            atl=atl,
            crt=crt,
            atlas_dir=path,
            target_size=100,
            max_failed_expansions=2,
            center=center,
            n_pubs_per_exp_max=10,
            call_size=None,
            n_sources_max=None,
            record_pubs_per_update=True,
        )

    def test_atlas_tracer_s2(self, tmp_path):
        path = tmp_path / atlas_dir
        path.mkdir()

        tracer = AtlasTracer(
            path,
            single_pub_bibtex_fp,
            "S2",
            "Word2Vec",
            vectorizer_kwargs={
                "corpus_path": astro_corpus_1,
                "model_path": model_path_1,
            },
        )
        tracer.expand_atlas(
            target_size=100,
            max_failed_expansions=2,
            n_pubs_max=10,
            call_size=None,
            n_sources_max=None,
            record_pubs_per_update=True,
        )

    def test_atlas_tracer_with_convergence_func(self, tmp_path):
        path = tmp_path / atlas_dir
        path.mkdir()

        tracer = AtlasTracer(
            path,
            single_pub_bibtex_fp,
            "S2",
            "Word2Vec",
            vectorizer_kwargs={
                "corpus_path": astro_corpus_1,
                "model_path": model_path_1,
            },
        )

        # Simple dummy convergence criterion: length is divisible by 3
        func = lambda atl: len(atl) % 3 == 0

        tracer.expand_atlas(
            target_size=100,
            max_failed_expansions=2,
            convergence_func=func,
            n_pubs_max=10,
            call_size=None,
            n_sources_max=None,
            record_pubs_per_update=True,
        )

    def test_atlas_tracer_ads(self, tmp_path):
        path = tmp_path / atlas_dir
        path.mkdir()

        tracer = AtlasTracer(
            path,
            single_pub_bibtex_fp,
            "ADS",
            "BOW",
            vectorizer_kwargs={
                "corpus_path": astro_corpus_1,
                "model_path": model_path_1,
            },
        )
        tracer.expand_atlas(
            target_size=100,
            max_failed_expansions=2,
            n_pubs_max=10,
            call_size=None,
            n_sources_max=None,
            record_pubs_per_update=True,
        )


class TestSearchConvergence:
    librarian = SemanticScholarLibrarian()
    vectorizer = SciBERTVectorizer()
    crt = Cartographer(librarian, vectorizer)

    # Construct Atlas
    bibtex_fp = ten_pub_bibtex_fp
    atl = crt.bibtex_to_atlas(bibtex_fp)
    atl = crt.project(atl)

    # Mock expansion/update history data
    input = [
        ["f2c251056dee4c6f9130b31e5e3e4b3296051c49"],  # it=0
        [
            "4364af31229f7e9a3d83a289a928b2f2a43d30cb",  # it=1
            "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
            "287fa946f30eaa78ea86f9c5bd61d67238202356",
        ],
        [
            "50dea78a96f03ba7fc3398c5deea5174630ef186",  # it=2
            "54a83cd1d94814b0f37ee48084260a2d1882648d",
            "4364af31229f7e9a3d83a289a928b2f2a43d30cb",
            "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
            "287fa946f30eaa78ea86f9c5bd61d67238202356",
            "2e6438be4901cb9b42ff23dcc3d433789b37d032",
            "04da6471743468b6bb1d26dd9a6eac4c03ca73ee",
        ],
        # it=3, len=10. This is equiv to atl.ids
        [
            "9d1a233164f27342d316662821e9a6bb855c25b4",
            "af6a1c9da102e29fee5d309ec33831207e9f23e5",
            # ----- it 2 ----
            "50dea78a96f03ba7fc3398c5deea5174630ef186",
            "54a83cd1d94814b0f37ee48084260a2d1882648d",
            "4364af31229f7e9a3d83a289a928b2f2a43d30cb",
            "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
            "287fa946f30eaa78ea86f9c5bd61d67238202356",
            "2e6438be4901cb9b42ff23dcc3d433789b37d032",
            "04da6471743468b6bb1d26dd9a6eac4c03ca73ee",
            # ----- end it 2 ----
            "0def4f553107451204b34470890d019b373798b5",
        ],
    ]

    crt.record_update_history(
        atl.ids,
        input,
    )

    kernel_size_arr = np.array(
        [
            [-1, -1, -1],
            [-1, -1, -1],
            [-1, -1, 0],
            [-1, -1, 1],
            [-1, 0, 3],
            [0, 0, 2],
            [-1, 0, 0],
            [-1, -1, 1],
            [-1, -1, 3],
            [-1, -1, -1],
        ]
    )

    # Simulates crt.track
    atl.history = {
        "pubs_per_update": input,
        "kernel_size": kernel_size_arr,
    }

    def test_search_converged_ids_final_column_nontrivial(self):
        atl = TestSearchConvergence.atl
        # Now compute the number of converged publications s.t. kernel_size > 2, starting at the update index over which 3 publications were added. This should give us the last index, since 3 were added then, and exactly 3 'converged' publications, since the last column of `expected` has 3 pubs with values >= 2

        # First test that we get the expected result when indexing
        # select the ids that have k=3,2, and 3 in last col of `expected`
        expected_ids = [x for i, x in enumerate(atl.ids) if i in [4, 5, 8]]

        actual_conv_ids = search_converged_ids(
            atl,
            num_pubs_added=3,
            kernel_size=2,
        )
        actual_sorted = sorted(actual_conv_ids)
        expected_sorted = sorted(expected_ids)

        assert actual_sorted == expected_sorted

    def test_search_converged_ids_final_column_trivial(self):
        # Now test we get the same thing with num_pubs added =0
        atl = TestSearchConvergence.atl
        expected_ids = [x for i, x in enumerate(atl.ids) if i in [4, 5, 8]]

        actual_conv_ids = search_converged_ids(
            atl,
            num_pubs_added=0,
            kernel_size=2,
        )
        actual_sorted = sorted(actual_conv_ids)
        expected_sorted = sorted(expected_ids)
        assert actual_sorted == expected_sorted

    def test_search_converged_ids_first_column_empty(self):
        atl = TestSearchConvergence.atl
        # Now test that when num_pubs_added is 100, this is greater than the total atlas size, so we should index the first update, and there will be 0 converged pubs, since kernel size = 2.
        expected_ids = []
        actual_conv_ids = search_converged_ids(
            atl,
            num_pubs_added=100,
            kernel_size=2,
        )
        actual_sorted = sorted(actual_conv_ids)
        expected_sorted = sorted(expected_ids)
        assert actual_sorted == expected_sorted

    def test_search_converged_ids_first_column_get_center(self):
        atl = TestSearchConvergence.atl
        # Now test the same condition, but only require kernel_size to be 0. Then we should still get one pub: the first pub added to the atlas.
        expected_ids = [
            "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
        ]
        actual_conv_ids = search_converged_ids(
            atl,
            num_pubs_added=100,
            kernel_size=0,
        )
        actual_sorted = sorted(actual_conv_ids)
        expected_sorted = sorted(expected_ids)
        assert actual_sorted == expected_sorted



"""Base class for vectorizing abstracts."""

import numpy as np
from abc import ABC, abstractmethod


class Vectorizer(ABC):
    @abstractmethod
    def embed_documents(
        self, docs: list[str], batch_size: int = 64
    ) -> dict[str, np.ndarray]:
        """Embed a list of documents into document vectors.

        Args:
            docs: the documents to embed.

            batch_size: the batch size to use.

        Returns:
            a dict of the form
            {
                "embeddings": a numpy array of shape `(num_successful, embedding_dim)`, containing the document embeddingss

                "success_indices": a numpy array of shape `(num_successful,)`, containing the indices of all the documents for which document embeddings were successfully obtained.

                "fail_indices": a numpy array of shape `(len(docs) - num_successful,)`, containing the indices of all the documents for which document embeddings could not be obtained
            }
            where the indices are with respect to the `docs` list passed.

        """
        raise NotImplementedError



"""Helper functions for analyzing data yielded by an atlas."""

import pandas as pd
import numpy as np

from ..mapping.atlas import Atlas
from ..mapping.cartography import Cartographer
from ..vectorization.vectorizer import Vectorizer


def atlas_to_measurements(
    atl: Atlas,
    vectorizer: Vectorizer,
    con_d: int,
    kernel_size=16,  # TODO: find a principled way of selecting this value., i.e. Fig. 5 from Imel & Hafen (2023). https://openreview.net/pdf?id=mISayy7DPI.
    metrics: list[str] = ["density", "edginess"],
    fields_of_study=None,
    max_year: int = 2023,  # consider 2022
) -> pd.DataFrame:
    """Compute the density, edginess, and citations per year metrics for each publicaation in an atlas w.r.t. a vectorizer and convergence configurations, and return the results in a dataframe.

    Args:
        atl: the Atlas to measure

        vectorizer: the Vectorizer to use to compute density and edginess

        con_d: a reverse index. This represents a convergence degree, in the sense that it is the number of updates before the last udpate to require that a publication's neighborhood has not changed identity of composition.

        This will be used to compute the inverse index for the second axis of the array `atl.history['kernel_size']`, representing the degree of convergence. For details about this array see `sciterra.mapping.cartography.Cartographer.converged_kernel_size`. Default is 1, which means we will filter to all publications that have not changed neighborhoods up to `kernel_size` up until the very last update. If 2, then up to the second to last update, etc.

        kernel_size: the minimum required size of the neighborhood that we will require to not have changed, w.r.t. `cond_d`. Default is 16.

        metrics: the list of str names corresponding to metrics to compute for the atlas. See `sciterra.mapping.topography` for possible metrics.

    """
    kernels = atl.history[
        "kernel_size"
    ]  # shape `(num_pubs, max_update)`, where `max_update` is typically the total number of updates if this function is called after the atlas has been sufficiently built out.
    max_update = kernels.shape[1]

    # Get all publications that have not changed neighborhoods up to kernel_size for the last con_d updates
    converged_filter = kernels[:, -con_d] >= kernel_size
    ids = np.array(atl.projection.index_to_identifier)
    converged_pub_ids = ids[converged_filter]

    print(
        f"Convergence degree {con_d} (out of {max_update} total updates) yields {len(converged_pub_ids)} ids out of {len(atl)} total ids."
    )

    # Optionally filter only to `field_of_study` publications
    if fields_of_study is not None:
        converged_pub_ids = [
            id
            for id in converged_pub_ids
            if any(fld in atl[id].fields_of_study for fld in atl[id].fields_of_study)
        ]

    # Compute density, edginess
    crt = Cartographer(
        vectorizer=vectorizer,
    )
    measurements = crt.measure_topography(
        atl,
        ids=converged_pub_ids,
        metrics=metrics,
        kernel_size=kernel_size,
    )

    # Get citations
    citations_per_year = [
        atl[id].citation_count / (max_year - atl[id].publication_date.year)
        if (
            atl[id].publication_date.year < max_year
            and atl[id].citation_count is not None
        )
        else 0.0
        for id in converged_pub_ids
    ]

    # Annotate the center (this feels inefficient, but oh well)
    is_center = [identifier == atl.center for identifier in converged_pub_ids]

    if not any(is_center):
        import warnings

        warnings.warn(
            f"The center publication is not within the set of convered publications."
        )

    df = pd.DataFrame(
        measurements,
        columns=metrics,
    )
    df["citations_per_year"] = citations_per_year
    df["is_center"] = is_center

    df = df[~np.isinf(df["density"])]  # drop infs which occur for BOW vectorizer
    # TODO what about other very high densities that result from close to 0?

    df.dropna(
        inplace=True,
    )

    print(f"There are {len(df)} total observations after filtering.")

    return df



"""The general container for data for any scientific publication, regardless of the API that was used to obtain it."""

import warnings
from datetime import date
from ..misc.utils import get_verbose, custom_formatwarning

warnings.formatwarning = custom_formatwarning


"""Things a publication must have.

1. identifier
2. abstract
3. references -- a list of publication identifiers
4. citations -- a list of publication identifiers
5. publication date
6. citation count

"""


# keys for data
FIELDS = [
    "identifier",
    "abstract",
    "publication_date",
    "citation_count",
    "citations",
    "references",
]

ADDITIONAL_FIELDS = [
    "doi",
    "url",
    "title",
    "issn",
]


class Publication:
    """The Publication is a standardized container a scientific publication's retrieved data.

    In general, all data-cleaning shoud be done prior to constructing a Publication, in order to keep the class minimal.

    Attributes:

        identifier:
            The string id that uniquely identifies the publication, used for
                - storing in an Atlas
                - querying an API

        abstract:
            The string corresponding to the publication's abstract

        publication_date:
            A datetime representing the date of publication

        citation_count:
            An int corresponding to the number of citations received by the publication
    """

    def __init__(self, data: dict, **kwargs) -> None:
        """Construct a publication.

        Args:
            data: to initialize attributes
        """
        # Below are the attributes we expect every publication to have. If a publication is missing these, it will be removed from analysis.
        self._identifier = None
        self._abstract = None
        self._publication_date = None
        self._citation_count = None
        self._fields_of_study = None

        # Regularize and store data, including but not limited to above attrs.
        self.init_attributes(data, **kwargs)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abstract(self) -> str:
        return self._abstract

    @property
    def publication_date(self) -> date:
        return self._publication_date

    @property
    def citations(self) -> list[str]:
        return self._citations

    @property
    def references(self) -> list[str]:
        return self._references

    @property
    def citation_count(self) -> int:
        """The citation_count can be different from the length of `citations`, since the number of citations listed for a paper might be different from the number of (valid) citing papers indexed on the relevant API."""
        return self._citation_count

    @property
    def fields_of_study(self) -> list[str]:
        return self._fields_of_study

    def __repr__(self) -> str:
        return "sciterra.publication.Publication:{}".format(self.identifier)

    def __str__(self) -> str:
        return self.identifier

    def __hash__(self) -> int:
        return hash(self.__dict__.values())

    def __eq__(self, __value: object) -> bool:
        return self.__dict__ == __value.__dict__

    def __lt__(self, __value: object) -> bool:
        return str(self) < str(__value)

    def init_attributes(self, data, **kwargs) -> None:
        verbose = get_verbose(kwargs)

        if "identifier" in data:
            val = data["identifier"]
            if not isinstance(val, str):
                raise ValueError
            self._identifier = val

        if "abstract" in data:
            val = data["abstract"]
            if not isinstance(val, str):
                raise ValueError
            self._abstract = val

        if "publication_date" in data:
            val = data["publication_date"]
            if not isinstance(val, date):
                raise ValueError
            self._publication_date = val

        if "citations" in data:
            val = data["citations"]
            if not isinstance(val, list):
                raise ValueError
            self._citations = val
        else:
            self._citations = []

        if "references" in data:
            val = data["references"]
            if not isinstance(val, list):
                raise ValueError
            self._references = val
        else:
            self._references = []

        if "citation_count" in data:
            val = data["citation_count"]
            if not isinstance(val, int):
                raise ValueError
            self._citation_count = val

        else:
            # we can use citations, but this is unexpected, so raise a warning.
            if self.citations and verbose:
                warnings.warn(
                    "Found an entry for 'citations' but no entry for citation_count; this is unexpected. Inferring value from citation_count."
                )
                self._citation_count = len(self.citations)

        if "fields_of_study" in data:
            val = data["fields_of_study"]
            if not isinstance(val, list):
                raise ValueError
            self._fields_of_study = val

        ######################################################################
        # Other attributes
        ######################################################################

        data_copy = {k: v for k, v in data.items() if k in ADDITIONAL_FIELDS}
        self.__dict__.update(data_copy)



"""Simple preprocessing of scientific abstracts prior to vectorization."""

import spacy

# Another off the shelf simple tokenizer
from gensim.utils import simple_preprocess


class CustomPreprocessor:
    def __init__(
        self,
        allowed_pos_tags: set = {"NOUN", "VERB", "ADJ"},
        model="en_core_web_sm",
    ) -> None:
        """Initialize a custom tokenizer.

        Args:

            allowed_pos_tags: keep and lemmatize words that are tagged as one of these POS categories.

            model: the name of the spacy language model to load, assuming it is already downloaded.
        """
        try:
            nlp = spacy.load(model)
        except OSError:
            raise OSError(
                f"Can't find model '{model}'; make sure you have run 'python3 -m spacy download {model}'!"
            )

        self.nlp = nlp
        self.allowed_pos_tags = allowed_pos_tags

    def custom_preprocess(
        self,
        document: str,
    ) -> list[str]:
        """Get all of the lemmas of the words in a document, filtering by POS.

        Args:
            document: a multi-sentence string

        Returns:
            a list of the lemmatized, filtered words in the document

        Given the domain-specificity, we choose to heuristically stem instead of performing full, linguistically precise lemmatization that would require detailed vocabulary rules. That said, the nltk WordNet lemmatizer doesn't immediately seem to do better than basic stemming

        See https://github.com/zhafen/cc/blob/master/cc/utils.py#L173.
        """
        return [
            token.lemma_
            for sent in self.nlp(document).sents
            for token in sent
            if token.pos_ in self.allowed_pos_tags
        ]



"""Functions for manipulating an atlas based on the document embeddings of the abstracts of its publications."""

import bibtexparser
import inspect
import warnings

import numpy as np

from . import topography
from .atlas import Atlas
from .publication import Publication
from ..librarians.librarian import Librarian
from ..vectorization.vectorizer import Vectorizer
from ..vectorization.projection import Projection, merge, get_empty_projection
from ..misc.utils import get_verbose, custom_formatwarning

from typing import Callable, Tuple
from tqdm import tqdm
from sklearn.metrics.pairwise import cosine_similarity

warnings.formatwarning = custom_formatwarning

##############################################################################
# Helper function
##############################################################################


def batch_cospsi_matrix(embeddings: np.ndarray) -> np.ndarray:
    """Batch-process a pairwise cosine similarity matrix between embeddings.

    In order to avoid memory errors (e.g. bus error, segfaults) resulting from too large arrays, we batch process the construction of the cospsi_matrix.

    Args:
        embeddings: a numpy array of embeddings of shape `(num_pubs, embedding_dim)`

    Returns:
        cosine_similarities: a 2D numpy array of shape `(num_pubs, num_pubs)` representing the pairwise cosine similarity between each embedding
    """
    batch_size = min(1000, len(embeddings))  # Define a batch size

    cosine_similarities = None
    print(
        f"computing cosine similarity for {len(embeddings)} embeddings with batch size {batch_size}."
    )
    for i in tqdm(range(0, len(embeddings), batch_size)):
        # Process batches to compute cosine similarity
        batch = embeddings[i : i + batch_size]
        if cosine_similarities is None:
            cosine_similarities = cosine_similarity(batch, embeddings)
        else:
            cosine_similarities = np.vstack(
                (cosine_similarities, cosine_similarity(batch, embeddings))
            )

    return cosine_similarities


# Helper function for filtering
def pub_has_attributes(
    pub: Publication,
    attributes: list[str],
) -> bool:
    """Return True if a publication has all `attributes`.

    Args:
        attributes: the list of attributes to check are not `None` for each publication from the atlas.
    """
    return pub is not None and all(
        [getattr(pub, attr) is not None for attr in attributes]
    )


def pub_has_fields_of_study(
    pub: Publication,
    fields_of_study: list[str],
) -> bool:
    """Return true if any of `pub.fields_of_study` are in passed `fields_of_study`."""
    return pub is not None and any(
        [field in fields_of_study for field in pub.fields_of_study]
    )


##############################################################################
# Cartographer
##############################################################################


class Cartographer:
    """A basic wrapper for obtaining and updating atlas projections.

    `self.librarian`: the Librarian object used to query a bibliographic database API.
    `self.vectorizer`: the Vectorizer object used to get a document embedding for each abstract
    `self.pubs_per_update`: a list of lists of publication str ids, representing the publications that exist at each time step / expansion update.
    `self.update_history`: an np.array of ints representing when publications were added. A value of -2 indicates no record of being added.
    """

    def __init__(
        self,
        librarian: Librarian = None,
        vectorizer: Vectorizer = None,
    ) -> None:
        self.librarian = librarian
        self.vectorizer = vectorizer

        self.pubs_per_update: list[list[str]] = []
        self.update_history: np.ndarray = None

    ######################################################################
    # Get an Atlas from bibtex
    ######################################################################

    def bibtex_to_atlas(self, bibtex_fp: str, *args, **kwargs) -> Atlas:
        """Convert a bibtex file to an atlas, by parsing each entry for an identifier, and querying an API for publications using `self.librarian`.

        NOTE: the identifiers in the corresponding atlas will be API-specific ids; there is no relationship between the parsed id used to query papers (e.g. 'DOI:XYZ' in the case of SemanticScholar) and the resulting identifier associated with the resulting Publication object, (a paperId, a bibcode, etc.) Therefore, the purpose of using the `bibtex_to_atlas` method is primarily for initializing literature exploration in a human-readable way. If you want to obtain as many publications as identifiers supplied, you need to use `get_publications`.

        Args:
            bibtex_fp: the filepath where the bibtex file is saved.

            args and kwargs are passed to `get_publications`.
        """
        verbose = get_verbose(kwargs)

        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        # Retrieve the identifier from each bibtex entry
        identifiers = [
            self.librarian.bibtex_entry_identifier(entry)
            for entry in bib_database.entries
        ]
        identifiers = [id for id in identifiers if id is not None]
        if len(identifiers) < len(bib_database.entries) and verbose:
            warnings.warn(
                f"Only obtained {len(identifiers)} publications out of {len(bib_database.entries)} total due to missing identifiers."
            )

        # Query
        results = self.librarian.get_publications(identifiers, *args, **kwargs)
        # Validate
        publications = [
            result
            for result in results
            if (
                result is not None
                and result.publication_date is not None
                and result.abstract is not None
                # identifier will never be none
            )
        ]
        if len(publications) < len(identifiers) and verbose:
            warnings.warn(
                f"Only obtained {len(publications)} publications out of {len(identifiers)} total due to querying-related errors or missing abstracts."
            )

        # Construct atlas
        atl = Atlas(publications)
        return atl

    ######################################################################
    # Project Atlas
    ######################################################################

    def project(self, atl: Atlas, **kwargs) -> Atlas:
        """Update an atlas with its projection, i.e. the document embeddings for all publications using `self.vectorizer`, removing publications with no abstracts.

        Args:
            atl: the Atlas containing publications to project to document embeddings

            kwargs: keyword arguments propagated to `filter_by_func`

        Returns:
            the updated atlas containing all nonempty-abstract-containing publications and their projection
        """
        verbose = get_verbose(kwargs)

        # Only project publications that have abstracts and publication dates
        atl_filtered = self.filter_by_func(atl, **kwargs)
        num_empty = len(atl) - len(atl_filtered)
        if num_empty and verbose:
            warnings.warn(
                f"{num_empty} publications were filtered due to missing crucial data or incorrect field of study. There are now {len(atl_filtered.bad_ids)} total ids that will be excluded in the future."
            )

        # Project
        embeddings = None
        # get only embeddings for publications not already projected in atlas
        previously_embedded_ids = []
        if atl_filtered.projection is not None:
            previously_embedded_ids = atl_filtered.projection.identifier_to_index
        embed_ids = [
            id for id in atl_filtered.publications if id not in previously_embedded_ids
        ]

        fail_ids = set()
        if embed_ids:
            if verbose:
                if atl_filtered.projection is not None:
                    warnings.warn(
                        f"Found {len(atl_filtered.publications) - len(atl_filtered.projection)} publications not contained in Atlas projection."
                    )
                warnings.warn(f"Embedding {len(embed_ids)} total documents.")

            # Embed documents
            result = self.vectorizer.embed_documents(
                [atl_filtered[id].abstract for id in embed_ids],
                batch_size=kwargs["batch_size"] if "batch_size" in kwargs else None,
            )
            embeddings = result["embeddings"]
            success_indices = result["success_indices"]
            fail_indices = result["fail_indices"]

            # get new set of bad ids
            atl_filtered.bad_ids = atl_filtered.bad_ids.union(fail_ids)

            if fail_indices.tolist() and verbose:
                warnings.warn(
                    f"Failed to get embeddings for all {len(embed_ids)} publications; only {len(embeddings)} will be added to the Atlas. There are now {len(atl_filtered.bad_ids)} total ids that will be excluded in the future."
                )

            embed_ids_array = np.array(embed_ids)
            success_ids = embed_ids_array[success_indices]
            fail_ids = set(embed_ids_array[fail_indices])

            # create new projection
            projection = Projection(
                identifier_to_index={
                    identifier: idx for idx, identifier in enumerate(success_ids)
                },
                index_to_identifier=tuple(success_ids),
                embeddings=embeddings,
            )

        if not embed_ids or embed_ids is None and verbose:
            warnings.warn(f"Obtained no new publication embeddings.")
            projection = get_empty_projection()

        # merge existing projection with new projection
        merged_projection = merge(atl_filtered.projection, projection)

        # prepare to overwrite atlas with publications corresponding to updated (merged) projection
        embedded_publications = {
            id: pub
            for id, pub in atl_filtered.publications.items()
            if id in merged_projection.identifier_to_index
        }

        # Overwrite atlas data
        atl_filtered.publications = embedded_publications
        atl_filtered.projection = merged_projection
        return atl_filtered

    ######################################################################
    # Sort Atlas
    ######################################################################

    def sort(
        self,
        atl: Atlas,
        center: str,
    ) -> Tuple[list[str], list[str]]:
        """Sort an atlas according to cosine similarity to a center publication.
        Like numpy argsort, this returns identifiers that can be used to
        index the original atlas.

        Args:
            atl: the atlas to sort

            center: center the search on this publication

        Returns:
            sorted_keys: keys in descending order of similarity to the center publication
            sorted_values: values in descending order of similarity to the center publication
        """

        # If atlas is initial
        if atl.projection is None:
            atl = self.project(atl)
            if atl.projection is None:
                raise Exception(
                    f"Initial projection of atlas failed; make sure the initial publication has all the required attributes."
                )

        if len(atl.projection):
            # build cosine similarity matrix, of shape (1, num_pubs)
            cospsi_matrix = cosine_similarity(
                atl.projection.identifiers_to_embeddings([center]),
                atl.projection.embeddings,
            )
            # get most similar keys from center, including center itself
            sort_inds = np.argsort(cospsi_matrix)[-1][::-1]
            # argsort orders from least to greatest similarity, so reverse
            sorted_keys = atl.projection.indices_to_identifiers(sort_inds)
            sorted_values = cospsi_matrix[0][sort_inds]

            return sorted_keys, sorted_values

    ######################################################################
    # Expand Atlas
    ######################################################################

    def expand(
        self,
        atl: Atlas,
        *args,
        center: str = None,
        n_pubs_max: int = 4000,
        n_sources_max: int = None,
        record_pubs_per_update: bool = False,
        **kwargs,
    ) -> Atlas:
        """Expand an atlas by retrieving a list of publications resulting from traversal of the citation network.

        Args:
            atl: the atlas containing the region to expand

            center: (if given) center the search on this publication, preferentially searching related publications. Default is `None`, and the expansion is not centered. To keep a consistent expansion around one center, you should pass `atl.center`.

            n_pubs_max: maximum number of publications allowed in the expansion.

            n_sources_max: maximum number of publications (already in the atlas) to draw references and citations from.

            record_pubs_per_update: whether to track all the publications that exist in the resulting atlas to `self.pubs_per_update`. This should only be set to `True` when you need to later filter by degree of convergence of the atlas.

        Returns:
            atl_expanded: the expanded atlas
        """

        # Get the keys to expand
        expand_keys = None
        if center is not None:
            expand_keys = self.sort(atl, center)[0]

        # If that didn't work, just use all the keys
        existing_keys = set(atl.ids)
        if expand_keys is None:
            expand_keys = existing_keys

        if n_sources_max is not None:
            expand_keys = expand_keys[:n_sources_max]

        # Get identifiers for the expansion
        # For each publication corresponding to an id in `expand_keys`, collect the ids corresponding to the publication's references and citations.
        ids = set()
        for key in expand_keys:
            ids_i = set(atl[key].references + atl[key].citations)
            # Prune for obvious overlap, and for ids that have previously failed
            ids.update(ids_i - existing_keys - atl.bad_ids)
            # Break when the search is centered and we're maxed out
            if len(ids) > n_pubs_max and center is not None:
                break
        ids = list(ids)

        if not ids:
            print("Overly-restrictive search, no ids to retrive.")

        # Sample to account for max number of publications we want to retrieve
        if len(ids) > n_pubs_max:
            ids = np.random.choice(ids, n_pubs_max, replace=False)

        print(f"Expansion will include {len(ids)} new publications.")

        # Retrieve publications from ids using a librarian
        new_publications = self.librarian.get_publications(ids, *args, **kwargs)

        # New atlas
        atl_exp = Atlas(new_publications)

        # Update the new atlas
        atl_exp.publications.update(atl.publications)
        atl_exp.bad_ids = atl.bad_ids
        atl_exp.projection = (
            atl.projection
        )  # new projection will be updated in `project`
        atl_exp.center = atl.center

        # Record the new list of publications
        if record_pubs_per_update:
            self.pubs_per_update.append(list(atl_exp.ids))

        return atl_exp

    ######################################################################
    # Filter Atlas
    ######################################################################

    def filter_by_func(
        self,
        atl: Atlas,
        require_func: Callable[[Publication], bool] = lambda pub: pub_has_attributes(
            pub,
            attributes=[
                "abstract",
                "publication_date",
                "fields_of_study",
            ],
        ),
        record_pubs_per_update=False,
        **kwargs,
    ) -> Atlas:
        """Update an atlas by dropping publications (and corresponding data in projection) when certain fields are empty.

        Args:
            atl: the Atlas containing publications to filter

            require_func: a function that takes a publication and returns True if it should be kept in the atlas. For example, if `func = lambda pub: pub_has_attributes(pub, ["abstract"])`, then all publications that have `None` for the attribute `abstract` will be removed from the atlas, along with the corresponding data in the projection.

            record_pubs_per_update: whether to track all the publications that exist in the resulting atlas to `self.pubs_per_update`. This should only be set to `True` when you need to later filter by degree of convergence of the atlas. This is an important parameter because `self.filter` is called in `self.project`, which typically is called after `self.expand`, where we pass in the same parameter.

        Returns:
            the filtered atlas
        """
        # Filter publications
        invalid_pubs = {
            id: pub for id, pub in atl.publications.items() if not require_func(pub)
        }

        # Do not update if unnecessary
        if not len(invalid_pubs):
            return atl

        atl_filtered = self.filter_by_ids(atl, drop_ids=invalid_pubs.keys())

        # Record only the publications in the history that weren't filtered out
        if record_pubs_per_update:
            self.pubs_per_update[-1] = atl_filtered.ids

        return atl_filtered

    def filter_by_ids(
        self,
        atl: Atlas,
        keep_ids: list[str] = None,
        drop_ids: list[str] = None,
    ) -> Atlas:
        """Update an atlas by dropping publications (and corresponding data in projection).

        Args:
            atl: the Atlas containing publications to filter

            keep_ids: the list of publication ids to NOT filter; all other publications in `atl` not matching one of these ids will be removed.

            drop_ids: the list of publications to filter; all publications in `atl` matching one of these ids will be removed.
        """

        if all(x is not None for x in [keep_ids, drop_ids]):
            raise ValueError(
                "You must pass exactly one of `keep_ids` or `drop_ids`, but both had a value that was not `None`."
            )
        if keep_ids is not None:
            filter_ids = set([id for id in atl.ids if id not in keep_ids])
        elif drop_ids is not None:
            filter_ids = set(drop_ids)
        else:
            raise ValueError(
                "You must pass exactly one of `keep_ids` or `drop_ids`, but both had value `None`."
            )

        # Keep track of the bad identifiers to skip them in future expansions
        new_bad_ids = atl.bad_ids.union(filter_ids)

        # Filter embeddings, ids from projection
        if atl.projection is None:
            new_projection = None
        else:
            filter_indices = set()
            idx_to_id_new = []
            # From indexing
            for idx, id in enumerate(atl.projection.index_to_identifier):
                if id in filter_ids:
                    filter_indices.add(idx)
                else:
                    idx_to_id_new.append(id)
            # From embeddings
            embeddings = np.array(
                [
                    embedding
                    for idx, embedding in enumerate(atl.projection.embeddings)
                    if idx not in filter_indices
                ]
            )
            # From identifier to index map
            id_to_idx_new = {id: idx for idx, id in enumerate(idx_to_id_new)}
            # Construct new, filtered projection
            new_projection = Projection(
                identifier_to_index=id_to_idx_new,
                index_to_identifier=idx_to_id_new,
                embeddings=embeddings,
            )

        # Keep only filtered publications
        new_publications = [
            pub for id, pub in atl.publications.items() if id not in filter_ids
        ]

        # Construct new atlas
        atl_filtered = Atlas(new_publications, new_projection)
        atl_filtered.bad_ids = new_bad_ids

        # If the center was filtered out, then reset center
        atl_filtered.center = atl.center
        if atl_filtered.center not in atl_filtered.publications:
            atl_filtered.center = None

        return atl_filtered

    ########################################################################
    # Record Atlas history
    ########################################################################

    def track(
        self,
        atl: Atlas,
        calculate_convergence: bool = False,
        pubs: list[str] = None,
        pubs_per_update: list[list[str]] = None,
    ) -> Atlas:
        """Overwrite the data associated with tracking degree of convergence of publications in an atlas over multiple expansions. N.B.: the atlas must be fully projected, or else `converged_kernel_size` will raise a KeyError. By default, this function will overwrite the `atl.history` with updated `self.pubs_per_update`, but not `kernel_size`, which requires computing the converged kernel size for every publication in the atlas.

        Args:
            atl: the Atlas that will be updated by overwriting `Atlas.history`

            calculate_convergence: whether to call `self.converged_kernel_size`, and store the results in the `atl.history`.

            pubs: the list of publications to pass to `self.record_update_history`. By default `None`, and the most recent list of publications in the `self.update_history` list will be used.

            pubs_per_update: the list of lists of publications to pass to `self.record_update_history`. By default `None`, adn the full history in `self.update_history` will be used.

        Returns:
            atl the updated Atlas
        """
        print("Tracking atlas...")
        self.record_update_history(pubs, pubs_per_update)
        # Skip expensive convergence calculation if possible
        kernel_size = self.converged_kernel_size(atl) if calculate_convergence else None
        atl.history = {
            "pubs_per_update": self.pubs_per_update
            if pubs_per_update is None
            else pubs_per_update,
            "kernel_size": kernel_size,
        }
        return atl

    ########################################################################
    # Record Atlas history
    ########################################################################

    def record_update_history(
        self,
        pubs: list[str] = None,
        pubs_per_update: list[list[str]] = None,
    ) -> None:
        """Record when publications were added, by updating atl.update_history.

        atl.update_history is a np.array of ints representing when publications were added. A value of -2 indicates no record of being added.

        Args:
            pubs: a list of str ids corresponding to publications at the final update in the update history. By default `None`, and `self.pubs_per_update[-1]` will be used.

            pubs_per_update: a list of which publications existed at which iteration, with the index of the overall list corresponding to the iteration the publication was added. By default `None`, and `self.pubs_per_update` will be used.

        Updates:
            `self.update_history`: an np.array of ints representing when publications were added. A value of -2 indicates no record of being added.

        Returns:
            `None`
        """
        if pubs is None:
            pubs = self.pubs_per_update[-1]

        if pubs_per_update is None:
            pubs_per_update = self.pubs_per_update

        # Loop backwards
        i_max = len(pubs_per_update) - 1
        update_history = np.full(len(pubs), -2)
        for i, pubs_i in enumerate(pubs_per_update[::-1]):
            is_in = np.isin(pubs, pubs_i)
            update_history[is_in] = i_max - i
        self.update_history = update_history

    ########################################################################
    # Calculate Atlas convergence
    ########################################################################

    def converged_kernel_size(self, atl: Atlas) -> np.ndarray:
        """Calculate the largest size of the kernel that's converged (at differing levels of convergence) for each publication in a sample at each update.

        Args:
            atl: Atlas containing publications; for each publication we compute the largest converged kernel size at each update

        Returns:
            kernel_size: an array of ints of shape `(num_pubs, max_update)` representing the kernel size for converged kernels.
                - The first column indicates the largest kernel size that hasn't changed since the beginning,
                - The second column indicates the largest kernel size that hasn't changed since the first update,
                - etc. for the nth column.
        """
        print("Calculating degree of convergence for all publications.")
        if self.update_history is None:
            raise ValueError(
                "update_history is None; make sure you have called record_update_history()!"
            )

        if -2 in self.update_history:
            raise ValueError(
                "Incomplete update history as indicated by entries with values of -2."
            )

        publications = np.array(atl.ids)

        # Compute pairwise cosine similarity, shape `(num_pubs,num_pubs)`
        cospsi_matrix = batch_cospsi_matrix(atl.projection.embeddings)

        # 0. Loop over each publication
        cospsi_kernel: list[list[int]] = []
        for pub in tqdm(publications, desc="calculating converged kernel size"):
            # 1. Identify the similarity with the other publications relative to this publication, and sort accordingly.
            cospsi = cospsi_matrix[publications == pub].flatten()  # shape `(num_pubs,)`
            sort_inds = np.argsort(cospsi)[::-1]  # shape `(num_pubs,)`

            # 2. Identify the expansion iteration at which those publications were added to the atlas (`sorted_history`).
            sorted_history = self.update_history[sort_inds]  # shape `(num_pubs,)`

            # 3. Identify the latest iteration at which any publication was added to the atlas; this can be less than the total iterations.
            last_update = self.update_history.max()

            # 4. Loop through each iteration until `last_update`, and identify which publications were added at or before that iteration.
            result = [
                # 5. Compute how many publications out we can go and still only contain publications added at or before that iteration.
                # Use `argmin` to get the first instance of False
                # Finally, subtract 1: we want the first index before False.
                # See tests.test_cartography.TestConvergence.test_converged_kernel_size for a concrete example.
                np.argmin(sorted_history <= update_idx) - 1
                for update_idx in range(last_update)
            ]  # shape `(last_update, )`

            cospsi_kernel.append(result)

        return np.array(cospsi_kernel)

    ########################################################################
    # Measure Atlas topography
    ########################################################################

    def measure_topography(
        self,
        atl: Atlas,
        ids: list[str] = None,
        metrics: list[str] = ["density"],
        min_prior_pubs: int = 2,
        kernel_size=16,
        **kwargs,
    ):
        """Measure topographic properties of all publications relative to prior
        publications.

        Args:

            atl: the Atlas to measure

            publication_indices: an np.ndarray of ints representing the indices of publications in the Atlas projection to measure

            metrics: A list of strings representing the metrics to use. Options are...
                constant_asymmetry: The asymmetry of a publication $p_i$ w.r.t the entire atlas $\\{ p_j \\forall j \\in \\{1, ..., k\\} \\} where $k$ is the length of the atlas

                    $| \\sum_{j}^{k-1}( p_i - p_j ) |$

                kernel_constant_asymmetry: The asymmetry of a publication w.r.t. its kernel, { p_j for all j in {1, ..., k} } where k is `kernel_size`, i.e. the k nearest neighbors.

                density: the density of a publication's surrounding area, estimated by a heuristic inspired by mass / volume = k publications divided by the minimum arc length enclosing the furthest publication.

                    $\\frac{ k }{ smoothing\\_length(k) }$

                smoothing_length: The distance (in radians) to the farthest publication in the kernel, i.e. the kth nearest neighbor.

            min_prior_pubs: The minimum number of publications prior to the target publication for which to calculate the metric.

            kernel_size: the number of publications surrounding the publication for which to compute the topography metric, i.e. k nearest neighbors for k=kernel_size.

        Returns:
            estimates: an np.ndarray of shape `(len(publication_indices), len(metrics))` representing the estimated topography metric values for each publication.
        """
        _ = get_verbose(kwargs)

        # By default calculate for all publications
        if ids is None:
            ids = atl.ids
        else:
            ids = list(ids)

        if not ids:
            raise Exception("No publications to measure topography of.")

        # Get publication dates, for filtering
        dates = np.array([atl[identifier].publication_date for identifier in ids])

        # Get pairwise cosine similarities for ids
        embeddings = atl.projection.identifiers_to_embeddings(ids)
        cospsi_matrix = batch_cospsi_matrix(embeddings)

        # From here on, use embedding indices instead of identifiers
        # our embeddings are already in the correct order, so just use them
        publication_indices = np.arange(len(embeddings))

        print(f"Computing {metrics} for {len(ids)} publications.")
        estimates = []
        for idx, identifier in tqdm(enumerate(ids), total=len(ids)):
            # Get the date of publication
            date = atl[identifier].publication_date

            # Identify prior publications
            is_prior = dates < date
            if is_prior.sum() < min_prior_pubs:
                estimates.append([np.nan for _ in metrics])
                continue

            # Choose valid publications
            is_other = publication_indices != idx
            is_valid = is_prior & is_other
            valid_indices = publication_indices[is_valid]

            kwargs = {
                "idx": idx,
                "cospsi_matrix": cospsi_matrix,
                "valid_indices": valid_indices,
                "publication_indices": publication_indices,
                "embeddings": embeddings,
                "kernel_size": kernel_size,
            }

            def call_metric(
                metric: str,
                **kwargs,
            ) -> float:
                """Wrapper function to simplify topography metric api."""
                # Get the metric
                fn = getattr(topography, f"{metric}_metric")

                # Identify arguments to pass
                fn_args = inspect.getfullargspec(fn)
                used_kwargs = {}
                for key, value in kwargs.items():
                    if key in fn_args.args:
                        used_kwargs[key] = value
                # Call
                estimate = fn(**used_kwargs)
                return estimate

            estimates.append([call_metric(metric, **kwargs) for metric in metrics])

        estimates = np.array(estimates)

        return estimates



"""Functions for measuring topographic properties of (the semantic feature space of publications inside) an Atlas."""

import numpy as np


########################################################################
# Density metrics
########################################################################


def smoothing_length_metric(
    idx: int,
    cospsi_matrix: np.ndarray,
    valid_indices: np.ndarray,
    kernel_size: int = 16,
):
    """Proxy for the density of a publication defined as the minimum
    arc length that encloses kernel_size other publications.

    Args:
        idx: the index of the vector to calculate the measurement for.

        cospsi_matrix: a 2D matrix of pairwise cosine similarity scores for publication embeddings.

        valid_indices: Indices of the other publication used when calculating the measurements.

        kernel_size: number of K nearest neighbors to calculate the measurement on.

    Returns:
        h: float representing arc length containing `kernel_size` other publications. (Assumes normalized to a radius of 1.)
    """

    # We can't have the kernel larger than the number of valid publications
    if kernel_size > len(valid_indices):
        return np.nan

    # Get 1D array of similarity scores to idx vector
    cospsi = cospsi_matrix[idx][valid_indices]

    # Get cosine distance to the least similar vector
    # np.sort orders from least to greatest similarity, so reverse after
    cospsi_max = np.sort(cospsi)[::-1][kernel_size - 1]

    # Compute arclength to furthest vector
    return np.arccos(cospsi_max)


def density_metric(
    idx: int,
    cospsi_matrix: np.ndarray,
    valid_indices: np.ndarray,
    kernel_size: int = 16,
):
    """Estimate the density of a publication by calculating the
    smoothing length that encloses kernel_size other publications.

    Args:
        idx: the index of the vector to calculate the measurement for.

        cospsi_matrix: a 2D matrix of pairwise cosine similarity scores for publication embeddings.

        valid_indices: Indices of the other publication used when calculating the measurements.

        kernel_size: number of K nearest neighbors to calculate the measurement on.

    Returns:
        density: a float representing `kernel_size` divided by arc length containing `kernel_size` other publications.
    """

    h = smoothing_length_metric(idx, cospsi_matrix, valid_indices, kernel_size)
    density = kernel_size / h

    # # TODO: there is serious numerical instability for BOW methods. Not sure what the most principled way to deal with them are.
    # if density > 1e7 and np.isfinite(density):
    #     breakpoint()

    return density


########################################################################
# Asymmetry metrics
########################################################################


def edginess_metric(
    idx: int,
    cospsi_matrix: np.ndarray,
    valid_indices: np.ndarray,
    publication_indices: np.ndarray,
    embeddings: np.ndarray,
    kernel_size: int = 16,
) -> float:
    """Estimate the asymmetry of a publication by calculating the difference between that publication's projection and the other publications within the kernel. Normalized to between 0 and 1.

    Args:
        idx: the index of the vector to calculate the measurement for.

        cospsi_matrix: an np.ndarray of shape `(num_pubs, num_pubs)` representing pairwise cosine similarity scores for publication embeddings.

        valid_indices: an np.ndarray of shape `(num_valid_pubs)` representing indices of the other publications used when calculating the measurements, i.e. num_valid_pubs <= num_pubs.

        publication_indices: an np.ndarray of shape `(num_pubs,)` representing indices of all publications in the atlas projection

        embeddings: an np.ndarray of shape `(num_pubs, embedding_dim)` vectors for all publications in the atlas projection

        kernel_size: number of K nearest neighbors to calculate the measurement on.

    Returns:
        a float representing the normalized magnitude of the asymmetry metric.

    """
    return (
        kernel_constant_asymmetry_metric(
            idx,
            cospsi_matrix,
            valid_indices,
            publication_indices,
            embeddings,
            kernel_size=kernel_size,
        )
        / kernel_size
    )


def kernel_constant_asymmetry_metric(
    idx: int,
    cospsi_matrix: np.ndarray,
    valid_indices: np.ndarray,
    publication_indices: np.ndarray,
    embeddings: np.ndarray,
    kernel_size: int = 16,
) -> float:
    """Estimate the asymmetry of a publication by calculating the difference
    between that publication's projection and the other publications within
    the kernel.

    Args:
        idx: an int representing the index of the vector to calculate the measurement for.

        cospsi_matrix: an np.ndarray of shape `(num_pubs, num_pubs)` representing pairwise cosine similarity scores for publication embeddings.

        valid_indices: an np.ndarray of shape `(num_valid_pubs)` representing indices of the other publications used when calculating the measurements, i.e. num_valid_pubs <= num_pubs.

        publication_indices: an np.ndarray of shape `(num_pubs,)` representing indices of all publications in the atlas projection

        embeddings: an np.ndarray of shape `(num_pubs, embedding_dim)` vectors for all publications in the atlas projection

        kernel_size: an int representing the number of K nearest neighbors to calculate the measurement on.

    Returns:
        mag: a float representing the magnitude of the asymmetry metric.
    """

    # We can't have the kernel larger than the number of valid publications
    if kernel_size > len(valid_indices):
        return np.nan

    # Input
    cospsi = cospsi_matrix[idx][valid_indices]
    sorted_inds = np.argsort(cospsi)[::-1][:kernel_size]
    other_inds = publication_indices[valid_indices][sorted_inds]
    embedding = embeddings[idx]
    other_embeddings = embeddings[other_inds]

    # Differences
    diff = embedding - other_embeddings
    diff_mag = np.linalg.norm(diff, axis=1)
    result = (diff / diff_mag[:, np.newaxis]).sum(axis=0)
    mag = np.linalg.norm(result)

    return mag



"""Classes for constructing maps of scientific literature.

The `sciterra.mapping.atlas` submodule contains the basic data structure, the Atlas.

The `sciterra.mapping.cartography` submodule contains functionality for manipulating an Atlas.

"""



"""
.. include:: ../../README.md
"""

__docformat__ = "google"

from .mapping.atlas import Atlas
from .mapping.cartography import Cartographer
from .mapping.publication import (
    Publication,
)  # publication should probably be moved out of mapping.



"""Test basic Atlas functionality, independently of API. To obtain realistic publication data, should probably read in a .bib file."""

import bibtexparser

import pandas as pd

from sciterra.mapping.atlas import Atlas
from sciterra.mapping.publication import Publication

from sciterra.misc.utils import write_pickle, read_pickle


atlas_empty_dir = "atlas_empty"
atlas_single_dir = "atlas_single"
atlas_10_dir = "atlas_10"


class TestAtlasDummy:
    def test_create_empty_atlas(self):
        atl = Atlas([])
        assert atl.publications == dict()

    def test_save_empty_atlas(self, tmp_path):
        path = tmp_path / atlas_empty_dir
        path.mkdir()
        atl = Atlas([])
        atl.save(path)

    def test_save_load_empty_atlas(self, tmp_path):
        path = tmp_path / atlas_empty_dir
        path.mkdir()
        embedding_path = path / "publications.pkl"
        write_pickle(embedding_path, [])

        atl = Atlas.load(path)
        assert atl.publications == Atlas([]).publications

    """Test an Atlas with a single publication."""

    def test_atlas_single(self):
        pub = Publication({"identifier": "id"})
        atl = Atlas([pub])
        assert atl[str(pub)] == pub

    def test_atlas_single_duplicate(self):
        pub = Publication({"identifier": "id"})
        atl = Atlas([pub, pub])
        assert atl[str(pub)] == pub
        assert len(atl) == 1

    def test_save_atlas_single(self, tmp_path):
        path = tmp_path / atlas_single_dir
        path.mkdir()
        atl = Atlas([Publication({"identifier": "id"})])
        atl.save(path)

    def test_save_load_atlas_single(self, tmp_path):
        path = tmp_path / atlas_single_dir
        path.mkdir()
        atl = Atlas([Publication({"identifier": "id"})])
        atl.save(path)

        atl_loaded = Atlas.load(path)
        assert atl.publications == atl_loaded.publications

    """Test an Atlas with 10 publications."""

    def test_atlas_10(self):
        pubs = [Publication({"identifier": f"id_{i}"}) for i in range(10)]
        atl = Atlas(pubs)
        for pub in pubs:
            assert atl[str(pub)] == pub

    def test_save_atlas_10(self, tmp_path):
        path = tmp_path / atlas_10_dir
        path.mkdir()

        pubs = [Publication({"identifier": f"id_{i}"}) for i in range(10)]
        atl = Atlas(pubs)
        atl.save(path)

    def test_save_load_atlas_10(self, tmp_path):
        path = tmp_path / atlas_10_dir
        path.mkdir()

        pubs = [Publication({"identifier": f"id_{i}"}) for i in range(10)]
        atl = Atlas(pubs)
        atl.save(path)

        atl_loaded = Atlas.load(path)
        assert atl.publications == atl_loaded.publications


class TestAtlasCitationNetwork:

    """Check the basic graph structure encoded into an Atlas by each publication's citations and references."""

    def test_singleton_network(self):
        pass

    def test_three_node_network(self):
        pass



"""GPT-2 is a large causal language model that achieved SOTA in many NLP tasks before its successors created by OpenAI.

Links:
    - Paper: https://cdn.openai.com/better-language-models/language_models_are_unsupervised_multitask_learners.pdf
    - HF: https://huggingface.co/docs/transformers/en/model_doc/gpt2
"""


# See https://huggingface.co/docs/transformers/en/model_doc/gpt2#transformers.GPT2Model

import torch
import numpy as np
from .vectorizer import Vectorizer
from tqdm import tqdm

from transformers import AutoTokenizer, GPT2Model, GPT2TokenizerFast


MPS_DEVICE = torch.device("mps")

# This is the hidden dimension size
EMBEDDING_DIM = 768

# Default is small, otherwise memory limits become a problem
BATCH_SIZE = 8


class GPT2Vectorizer(Vectorizer):
    def __init__(self, device="cuda", **kwargs) -> None:
        # Get tokenizer
        self.tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")

        # Need to specify padding token
        self.tokenizer.pad_token = self.tokenizer.eos_token

        # Get model
        self.model = GPT2Model.from_pretrained("gpt2")

        # set device to GPU
        if device == "mps":
            self.device = MPS_DEVICE
        elif device == "cuda":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        print(f"Using device: {self.device}.")
        self.model.to(self.device)

        # Put the model in "evaluation" mode
        self.model.eval()

        super().__init__()

    def embed_documents(
        self,
        docs: list[str],
        batch_size: int = BATCH_SIZE,
    ) -> dict[str, np.ndarray]:
        """Embed a list of documents (raw text) into GPT-2 vectors, by batching.

        Args:
            docs: the documents to embed.

        Returns:
            a numpy array of shape `(num_documents, embedding_dim)`

        """
        if batch_size is None:
            batch_size = BATCH_SIZE

        embeddings = []

        pbar = tqdm(
            total=len(docs),
            desc="embedding documents",
            leave=True,
        )

        for i in range(0, len(docs), batch_size):
            batch = docs[i : min(len(docs), i + batch_size)]

            # A note on padding from Thomas Wolf: https://github.com/huggingface/transformers/issues/808#issuecomment-522932583

            # Tokenize the batch
            encoded = self.tokenizer(
                batch,
                add_special_tokens=True,
                padding=True,  # pad up to length of longest abstract
                truncation=True,  # max length 1024 tokens
                return_tensors="pt",
            )
            # each encoded item of shape [64, 1024]
            input_ids = encoded["input_ids"]
            assert input_ids.size()[-1] <= 1024

            # Put data on GPU
            for k, v in encoded.items():
                encoded[k] = v.to(self.device)

            # Run the text through GPT2,
            # collecting all of the hidden states produced from all 12 layers.
            with torch.no_grad():
                outputs = self.model(  # discard logits
                    **encoded,
                )

            # Get the embeddings of each final token in the batch

            # Get the full last hidden state,
            # shape [batch_size, sequence_length, hidden_size=768]
            last_hidden_state = outputs.last_hidden_state

            # Get the varying sequence lengths,
            # shape [batch_size,]
            sequence_lengths = torch.tensor(
                [
                    torch.nonzero(token_ids.eq(self.tokenizer.pad_token_id))[0].item()
                    + 1
                    if token_ids.eq(self.tokenizer.pad_token_id).any()
                    else len(token_ids)
                    for token_ids in input_ids
                ]
            )

            # Get embeddings of each final token,
            # shape [batch_size, hidden_size]
            last_hidden_states = torch.stack(
                [
                    last_hidden_state[i, sequence_lengths[i] - 1, :]
                    for i in range(len(sequence_lengths))
                ]
            )

            # Move to the CPU and convert to numpy ndarray
            batched_embeddings = last_hidden_states.detach().cpu().numpy()

            # Collect batched embeddings
            embeddings.extend(batched_embeddings)

            pbar.update(batch_size)
        pbar.close()

        # We don't deal with OOV, so we always return full list of ids
        return {
            "embeddings": np.array(embeddings),
            "success_indices": np.arange(len(embeddings), dtype=int),
            "fail_indices": np.array([], dtype=int),
        }



"""Test basic expansion functionality with Cartographer, which has some convenient wrappers around each API librarian."""

import bibtexparser

import numpy as np

from datetime import datetime

from sciterra.mapping.atlas import Atlas
from sciterra.mapping.cartography import (
    Cartographer,
    pub_has_attributes,
)
from sciterra.librarians.s2librarian import SemanticScholarLibrarian
from sciterra.mapping.publication import Publication
from sciterra.vectorization import SciBERTVectorizer, Word2VecVectorizer

bib_dir = "src/tests/data/bib"
single_pub_bibtex_fp = f"{bib_dir}/single_publication.bib"
ten_pub_bibtex_fp = f"{bib_dir}/ten_publications.bib"
realistic_bibtex_fp = f"{bib_dir}/rdsg.bib"
corpus_path = "src/tests/data/corpora/astro_1.txt"

##############################################################################
# SemanticScholar x SciBERT
##############################################################################

atlas_dir = "atlas_tmpdir"

# NOTE: Any time you are querying an API for papers, it is not a good idea to have strict tests on the resulting output size, since there is a significant amount of things out of our control, including that the online database may have added new papers.


class TestS2BibtexToAtlas:
    """Test functionality that maps a bibtex file to a list of identifiers, and then populates an atlas. The purpose of this functionality is to map a human-readable / very popular dataformat to the Atlas data structure."""

    librarian = SemanticScholarLibrarian()
    crt = Cartographer(librarian)

    def test_bibtex_to_atlas_single(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)
        entry: dict = bib_database.entries[0]

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBProjection.crt.bibtex_to_atlas(bibtex_fp)

        (pub,) = list(atl.publications.values())

        assert pub.identifier
        assert pub.abstract
        assert pub.publication_date
        assert pub.citation_count >= 0
        assert len(pub.citations) >= 0
        assert len(pub.references) >= 0

        assert entry["doi"] == pub.doi

    def test_bibtex_to_atlas_ten(self, tmp_path):
        # Load ten files from bibtex
        # Load expected values
        bibtex_fp = ten_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        dois = [entry["doi"] for entry in bib_database.entries if "doi" in entry]

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBProjection.crt.bibtex_to_atlas(bibtex_fp)

        for id, pub in atl.publications.items():
            assert pub.identifier == id
            assert pub.abstract
            assert pub.publication_date
            assert pub.citation_count >= 0
            assert len(pub.citations) >= 0
            assert len(pub.references) >= 0
            assert pub.doi in dois if hasattr(pub, "doi") else True

    def test_bibtex_to_atlas_realistic(self, tmp_path):
        # Load ten files from bibtex
        # Load expected values
        bibtex_fp = realistic_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBProjection.crt.bibtex_to_atlas(
            bibtex_fp,
            # multiprocess=False,
        )

        for id, pub in atl.publications.items():
            assert pub.identifier == id
            assert pub.abstract
            assert pub.publication_date
            assert pub.citation_count >= 0
            assert len(pub.citations) >= 0
            assert len(pub.references) >= 0

        # I find that I get 28 out of 86 total refs, i.e. less than a third of papers targeted.
        # or 32 lol
        # or 31
        # assert len(atl) == 28
        # assert len(atl) == 32
        # assert len(atl) == 31


class TestS2SBProjection:
    librarian = SemanticScholarLibrarian()
    vectorizer = SciBERTVectorizer()
    crt = Cartographer(librarian, vectorizer)

    def test_empty_projection(self):
        pubs = [Publication({"identifier": f"id_{i}"}) for i in range(10)]
        atl = Atlas(pubs)

        atl_proj = TestS2SBProjection.crt.project(atl)
        assert atl_proj.projection is None  # was filtered

    def test_dummy_projection_no_date(self):
        pubs = [
            Publication({"identifier": f"id_{i}", "abstract": "blah blah blah"})
            for i in range(10)
        ]
        atl = Atlas(pubs)

        atl_proj = TestS2SBProjection.crt.project(atl)
        assert all([hasattr(pub, "abstract") for pub in atl.publications.values()])
        assert atl_proj.projection is None  # was filtered

    def test_dummy_projection_no_abstract(self):
        pubs = [
            Publication(
                {"identifier": f"id_{i}", "publication_date": datetime(2023, 1, 1)}
            )
            for i in range(10)
        ]
        atl = Atlas(pubs)

        # breakpoint()
        atl_proj = TestS2SBProjection.crt.project(atl)

        assert all(
            [hasattr(pub, "publication_date") for pub in atl.publications.values()]
        )
        assert atl_proj.projection is None  # was filtered

    def test_dummy_projection(self):
        pubs = [
            Publication(
                {
                    "identifier": f"id_{i}",
                    "abstract": "blah blah blah",
                    "publication_date": datetime(2023, 1, 1),
                    "fields_of_study": ["dummy_field"],
                }
            )
            for i in range(10)
        ]
        atl = Atlas(pubs)

        atl_proj = TestS2SBProjection.crt.project(atl)

        projection = atl_proj.projection

        vector0 = projection.identifiers_to_embeddings(["id_0"])
        vector1 = projection.identifiers_to_embeddings(["id_9"])
        assert np.array_equal(vector0, vector1)

    def test_dummy_projection_partial(self):
        crt = Cartographer(vectorizer=Word2VecVectorizer(corpus_path=corpus_path))

        pubs = [
            Publication(
                {
                    "identifier": f"id_{0}",
                    "abstract": "We use cosmological hydrodynamic simulations with stellar feedback from the FIRE (Feedback In Realistic Environments) project to study the physical nature of Lyman limit systems (LLSs) at z ≤ 1.",  # everything here should be in the Word2Vec default vocab, since it trains on this abstract.
                    "publication_date": datetime(2023, 1, 1),
                    "fields_of_study": ["dummy_field"],
                }
            ),
            Publication(
                {
                    "identifier": f"id_{1}",
                    "abstract": "outofvocabularyitem",  # this should not
                    "publication_date": datetime(2023, 1, 1),
                }
            ),
            Publication(
                {
                    "identifier": f"id_{2}",
                    "abstract": "We use cosmological hydrodynamic simulations with stellar feedback from the FIRE (Feedback In Realistic Environments) project to study the physical nature of Lyman limit systems (LLSs) at z ≤ 1.",
                    "publication_date": datetime(2023, 1, 1),
                    "fields_of_study": ["dummy_field"],
                }
            ),
        ]
        atl = Atlas(pubs)

        atl_proj = crt.project(atl)

        assert len(atl_proj) == 2

    def test_single_projection(self, tmp_path):
        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBProjection.crt.bibtex_to_atlas(single_pub_bibtex_fp)

        atl_proj = TestS2SBProjection.crt.project(atl)
        projection = atl_proj.projection

        identifier = atl.ids[0]
        assert projection.identifier_to_index == {identifier: 0}
        assert projection.index_to_identifier == (identifier,)
        assert projection.embeddings.shape == (1, 768)  # (num_pubs, embedding_dim)

    def test_project_correct_number(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBProjection.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        ids = pub.citations + pub.references
        center = pub.identifier

        atl_exp_single = TestS2SBProjection.crt.expand(atl, center=center)
        atl_exp_single = TestS2SBProjection.crt.project(atl_exp_single)

        before = len(atl_exp_single)
        atl_exp_double = TestS2SBProjection.crt.expand(
            atl_exp_single, center=center, n_pubs_max=200
        )
        after = len(atl_exp_double)

        # Check that the second projection does not need to pull more docs than necessary

        # 1. Simulate first part of project
        # 'only project publications that have abstracts'
        atl_filtered = TestS2SBProjection.crt.filter_by_func(
            atl_exp_double,
            require_func=lambda pub: pub_has_attributes(
                pub,
                attributes=["abstract"],
            ),
        )

        # 'get only embeddings for publications not already projected in atlas'
        previously_embedded_ids = []
        if atl_filtered.projection is not None:
            previously_embedded_ids = atl_filtered.projection.identifier_to_index
        embed_ids = [
            id for id in atl_filtered.publications if id not in previously_embedded_ids
        ]

        # 2. Check that the number of abstracts to be embedded does not exceed the size of the previous expansion
        assert len(embed_ids) <= after - before


class TestS2SBSort:
    librarian = SemanticScholarLibrarian()
    vectorizer = SciBERTVectorizer()
    crt = Cartographer(librarian, vectorizer)

    def test_argsort(self, tmp_path):
        # TODO: This takes a while, and we can probably reduce the time

        # Load single file from bibtex
        # Load expected values
        bibtex_fp = ten_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBExpand.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        center = pub.identifier

        sorted_keys, sorted_values = TestS2SBSort.crt.sort(atl, center=center)
        assert len(sorted_keys) == 10
        assert sorted_keys[0] == center
        assert sorted_values[0] > sorted_values[1]
        assert sorted_values[1] > sorted_values[-1]


class TestS2SBExpand:
    librarian = SemanticScholarLibrarian()
    vectorizer = SciBERTVectorizer()
    crt = Cartographer(librarian, vectorizer)

    def test_expand_single(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBExpand.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        ids = pub.citations + pub.references

        atl_exp = TestS2SBExpand.crt.expand(atl)

        assert len(atl_exp) > len(atl)
        # so far this holds, but things that aren't our fault could make it fail.
        assert len(atl_exp) == len(ids)

    def test_expand_double(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBExpand.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        ids = pub.citations + pub.references

        atl_exp_single = TestS2SBExpand.crt.expand(atl)
        atl_exp_double = TestS2SBExpand.crt.expand(atl_exp_single, n_pubs_max=200)
        # empirically found this
        # note that all ids from atl_exp_single is 68282!
        assert len(atl_exp_double) == 200 + len(ids)

        # Save atlas
        atl_exp_double.save(path)

    def test_expand_center_single(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBExpand.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        ids = pub.citations + pub.references
        center = pub.identifier

        atl_exp_single = TestS2SBExpand.crt.expand(atl, center=center)
        assert len(atl_exp_single) == len(ids)

        # Save atlas
        atl_exp_single.save(path)

    def test_expand_center_double(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestS2SBExpand.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        ids = pub.citations + pub.references
        center = pub.identifier

        atl_exp_single = TestS2SBExpand.crt.expand(atl, center=center)
        atl_exp_single = TestS2SBExpand.crt.project(atl_exp_single)
        atl_exp_double = TestS2SBExpand.crt.expand(
            atl_exp_single, center=center, n_pubs_max=200
        )
        # empirically found this
        # do no assert len(atl_exp_double)  == 4000 + len(ids), because we want 4000 + len(valid_ids), i.e. 148 not 154
        # assert len(atl_exp_double) == 348 # why off by a few?
        # assert len(atl_exp_double) == 345

        atl_exp_double.save(path)


class TestTopography:
    librarian = SemanticScholarLibrarian()
    vectorizer = SciBERTVectorizer()
    crt = Cartographer(librarian, vectorizer)

    def test_measure_topography_full(self):
        bibtex_fp = ten_pub_bibtex_fp
        atl = TestTopography.crt.bibtex_to_atlas(bibtex_fp)
        atl = TestTopography.crt.project(atl)
        metrics = [
            "density",
            "edginess",
        ]
        measurements = TestTopography.crt.measure_topography(
            atl,
            metrics=metrics,
        )
        assert measurements.shape == tuple((len(atl), len(metrics)))

    def test_measure_topography_subset(self):
        bibtex_fp = ten_pub_bibtex_fp
        atl = TestTopography.crt.bibtex_to_atlas(bibtex_fp)
        atl = TestTopography.crt.project(atl)
        ids = atl.ids[:-5]
        metrics = [
            "density",
            "edginess",
        ]
        measurements = TestTopography.crt.measure_topography(
            atl,
            ids=ids,
            metrics=metrics,
        )
        assert measurements.shape == tuple((len(ids), len(metrics)))

    def test_measure_topography_realistic(self):
        # Load single file from bibtex
        bibtex_fp = single_pub_bibtex_fp

        # Construct Atlas
        atl = TestTopography.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        ids = pub.citations + pub.references
        center = pub.identifier

        # Expand
        atl_exp_single = TestTopography.crt.expand(
            atl,
            center=center,
            n_pubs_max=200,
        )

        # Project, necessary for metrics!
        atl_exp_single = TestTopography.crt.project(atl_exp_single)
        ids = atl_exp_single.ids

        metrics = [
            "density",
            "edginess",
        ]
        measurements = TestTopography.crt.measure_topography(
            atl_exp_single,
            ids=ids,
            metrics=metrics,
        )
        assert measurements.shape == tuple((len(ids), len(metrics)))


class TestConvergence:
    librarian = SemanticScholarLibrarian()
    vectorizer = SciBERTVectorizer()
    crt = Cartographer(librarian, vectorizer)

    def test_record_update_history(self):
        # Construct Atlas
        bibtex_fp = ten_pub_bibtex_fp

        atl = TestConvergence.crt.bibtex_to_atlas(bibtex_fp)

        # Mock expansion/update history data
        history = [
            ["f2c251056dee4c6f9130b31e5e3e4b3296051c49"],  # it=0
            [
                "4364af31229f7e9a3d83a289a928b2f2a43d30cb",  # it=1
                "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
                "287fa946f30eaa78ea86f9c5bd61d67238202356",
            ],
            [
                "50dea78a96f03ba7fc3398c5deea5174630ef186",  # it=2
                "54a83cd1d94814b0f37ee48084260a2d1882648d",
                "4364af31229f7e9a3d83a289a928b2f2a43d30cb",
                "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
                "287fa946f30eaa78ea86f9c5bd61d67238202356",
                "2e6438be4901cb9b42ff23dcc3d433789b37d032",
                "04da6471743468b6bb1d26dd9a6eac4c03ca73ee",
            ],
        ]

        TestConvergence.crt.record_update_history(
            atl.ids,
            pubs_per_update=history,
        )

        expected = np.array(
            [
                -2,
                -2,
                2,
                2,
                1,
                0,
                1,
                2,
                2,
                -2,
            ]
        )
        actual = TestConvergence.crt.update_history

        assert np.array_equal(expected, actual)

    def test_converged_kernel_size(self):
        # Construct Atlas
        bibtex_fp = ten_pub_bibtex_fp
        atl = TestConvergence.crt.bibtex_to_atlas(bibtex_fp)
        atl = TestConvergence.crt.project(atl)

        # Mock expansion/update history data
        input = [
            ["f2c251056dee4c6f9130b31e5e3e4b3296051c49"],  # it=0
            [
                "4364af31229f7e9a3d83a289a928b2f2a43d30cb",  # it=1
                "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
                "287fa946f30eaa78ea86f9c5bd61d67238202356",
            ],
            [
                "50dea78a96f03ba7fc3398c5deea5174630ef186",  # it=2
                "54a83cd1d94814b0f37ee48084260a2d1882648d",
                "4364af31229f7e9a3d83a289a928b2f2a43d30cb",
                "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
                "287fa946f30eaa78ea86f9c5bd61d67238202356",
                "2e6438be4901cb9b42ff23dcc3d433789b37d032",
                "04da6471743468b6bb1d26dd9a6eac4c03ca73ee",
            ],
            # it=3, len=10. This is equiv to atl.ids
            [
                "9d1a233164f27342d316662821e9a6bb855c25b4",
                "af6a1c9da102e29fee5d309ec33831207e9f23e5",
                # ----- it 2 ----
                "50dea78a96f03ba7fc3398c5deea5174630ef186",
                "54a83cd1d94814b0f37ee48084260a2d1882648d",
                "4364af31229f7e9a3d83a289a928b2f2a43d30cb",
                "f2c251056dee4c6f9130b31e5e3e4b3296051c49",
                "287fa946f30eaa78ea86f9c5bd61d67238202356",
                "2e6438be4901cb9b42ff23dcc3d433789b37d032",
                "04da6471743468b6bb1d26dd9a6eac4c03ca73ee",
                # ----- end it 2 ----
                "0def4f553107451204b34470890d019b373798b5",
            ],
        ]

        TestConvergence.crt.record_update_history(
            atl.ids,
            input,
        )

        expected = np.array(
            [
                3,
                3,
                2,
                2,
                1,
                0,
                1,
                2,
                2,
                3,
            ]
        )
        actual = TestConvergence.crt.update_history

        assert np.array_equal(expected, actual)
        # mock center
        actual = TestConvergence.crt.converged_kernel_size(atl)

        expected = np.array(
            [
                [-1, -1, -1],
                [-1, -1, -1],
                [-1, -1, 0],
                [-1, -1, 1],
                [-1, 0, 3],
                [0, 0, 2],
                [-1, 0, 0],
                [-1, -1, 1],
                [-1, -1, 3],
                [-1, -1, -1],
            ]
        )

        assert np.array_equal(actual, expected)

    def test_pubs_per_update_expand_consistent(self, tmp_path):
        # Load single file from bibtex
        # Load expected values
        bibtex_fp = single_pub_bibtex_fp
        with open(bibtex_fp, "r") as f:
            bib_database = bibtexparser.load(f)

        path = tmp_path / atlas_dir
        path.mkdir()
        # Construct Atlas
        atl = TestConvergence.crt.bibtex_to_atlas(bibtex_fp)

        pub = list(atl.publications.values())[0]
        center = pub.identifier

        # Expand repeatedly
        num_expansions = 10
        for _ in range(num_expansions):
            atl = TestConvergence.crt.expand(
                atl,
                center=center,
                n_pubs_max=10,
                record_pubs_per_update=True,
            )

        assert len(TestConvergence.crt.pubs_per_update) == num_expansions

        TestConvergence.crt.record_update_history()

        # need to project all pubs before kernel calculations!
        atl = TestConvergence.crt.project(atl)

        # test convergence calculations
        result = TestConvergence.crt.converged_kernel_size(atl)



"""SciBERT is a BERT model trained on scientific text.

Links:
    Paper: https://aclanthology.org/D19-1371/
    Github:  https://github.com/allenai/scibert
    HF: https://huggingface.co/allenai/scibert_scivocab_uncased
"""

import torch
import numpy as np
from .vectorizer import Vectorizer
from tqdm import tqdm

from transformers import BertTokenizerFast, AutoModelForSequenceClassification
from transformers import logging

logging.set_verbosity(logging.ERROR)  # Silence warnings about training SCIBERT

MPS_DEVICE = torch.device("mps")

# the SciBERT pretrained model path from Allen AI repo
MODEL_PATH = "allenai/scibert_scivocab_uncased"
EMBEDDING_DIM = 768

BATCH_SIZE = 64


class SciBERTVectorizer(Vectorizer):
    def __init__(self, device="cuda", **kwargs) -> None:
        # Get tokenizer
        # TODO: does this include the SCIVOCAB or BASEVOCAB?
        self.tokenizer = BertTokenizerFast.from_pretrained(
            MODEL_PATH,
            do_lower_case=True,
            model_max_length=512,  # I shouldn't have to pass this but I do
        )
        # Get the model
        self.model = AutoModelForSequenceClassification.from_pretrained(
            pretrained_model_name_or_path=MODEL_PATH,
            output_attentions=False,
            output_hidden_states=True,
        )

        # set device to GPU
        if device == "mps":
            self.device = MPS_DEVICE
        elif device == "cuda":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        print(f"Using device: {self.device}.")
        self.model.to(self.device)

        # Put the model in "evaluation" mode
        self.model.eval()
        super().__init__()

    def embed_documents(
        self, docs: list[str], batch_size: int = BATCH_SIZE
    ) -> dict[str, np.ndarray]:
        """Embed a list of documents (raw text) into SciBERT vectors, by batching.

        Args:
            docs: the documents to embed.

        Returns:
            a numpy array of shape `(num_documents, 768)`

        """
        if batch_size is None:
            batch_size = BATCH_SIZE

        embeddings = []

        pbar = tqdm(
            total=len(docs),
            desc="embedding documents",
            leave=True,
        )

        for i in range(0, len(docs), batch_size):
            batch = docs[i : min(len(docs), i + batch_size)]

            # Tokenize the batch
            encoded = self.tokenizer(
                batch,
                add_special_tokens=True,
                padding=True,  # pad up to length of longest abstract
                truncation=True,  # max length 512 chars, unfortunately
                return_tensors="pt",
            )
            # each encoded item of shape [64, 512]
            assert encoded["input_ids"].size()[-1] <= 512

            # Put data on GPU
            for k, v in encoded.items():
                encoded[k] = v.to(self.device)

            # Run the text through SciBERT,
            # collecting all of the hidden states produced from all 12 layers.
            with torch.no_grad():
                _, encoded_layers = self.model(  # discard logits
                    **encoded,
                    return_dict=False,
                )

            # Extract the embeddings
            # index last (13th) BERT layer before the classifier
            final_hidden_state = encoded_layers[12]  # [batch_size, seq_len, 768]
            # index first token of sequence, [CLS], for our document embeddings
            batched_embeddings = final_hidden_state[:, 0, :]  # [batch_size, 768]

            # Move to the CPU and convert to numpy ndarray
            batched_embeddings = batched_embeddings.detach().cpu().numpy()

            # Collect batched embeddings
            embeddings.extend(batched_embeddings)

            pbar.update(batch_size)
        pbar.close()

        # We don't deal with OOV, so we always return full list of ids
        return {
            "embeddings": np.array(embeddings),
            "success_indices": np.arange(len(embeddings), dtype=int),
            "fail_indices": np.array([], dtype=int),
        }



"""Miscellaneous helper functions."""

import pickle
import time
from functools import wraps
from requests.exceptions import ReadTimeout, ConnectionError

# For Bibtex parsing


def standardize_month(month: str) -> str:
    month = month.lower()

    if len(month) == 3:
        standardized = {
            "jan": "January",
            "feb": "February",
            "mar": "March",
            "apr": "April",
            "may": "May",
            "jun": "June",
            "jul": "July",
            "aug": "August",
            "sep": "September",
            "oct": "October",
            "nov": "November",
            "dec": "December",
        }[month]
    else:
        standardized = {
            "january": "January",
            "february": "February",
            "march": "March",
            "april": "April",
            "may": "May",
            "june": "June",
            "july": "July",
            "august": "August",
            "september": "September",
            "october": "October",
            "november": "November",
            "december": "December",
        }[month]

    return standardized


# For querying apis


def keep_trying(
    n_attempts=5,
    allowed_exceptions=(ReadTimeout, ConnectionError),
    verbose=True,
    sleep_after_attempt=1,
):
    """Sometimes we receive server errors. We don't want that to disrupt the entire process, so this decorator allow trying n_attempts times.

    ## API_extension::get_data_via_api
    ## This decorator is general, except for the default allowed exception.

    Args:
        n_attempts (int):
            Number of attempts before letting the exception happen.

        allowed_exceptions (tuple of class):
            Allowed exception class. Set to BaseException to keep trying regardless of exception.

        sleep_after_attempt (int):
            Number of seconds to wait before trying each additional attempt.

        verbose (bool):
            If True, be talkative.

    Example Usage:
        > @keep_trying( n_attempts=4 )
        > def try_to_call_web_api():
        >     " do stuff "
    """

    def _keep_trying(f):
        @wraps(f)
        def wrapped_fn(*args, **kwargs):
            # Loop over for n-1 attempts, trying to return
            for i in range(n_attempts - 1):
                # waiting may help with connection errors?
                time.sleep(sleep_after_attempt)

                try:
                    result = f(*args, **kwargs)
                    if i > 0 and verbose:
                        print(f"Had to call {f} {i+1} times to get a response.")
                    return result
                # TODO: need to branch for every case.
                except allowed_exceptions as _:
                    continue

            # On last attempt just let it be
            if verbose:
                print(
                    f"Had to call {f} {n_attempts} times to get a response. Trying once more."
                )
            return f(*args, **kwargs)

        return wrapped_fn

    return _keep_trying


def chunk_ids(ids: list[str], call_size):
    """Helper function to chunk bibcodes or paperIds into smaller sublists if appropriate."""
    # Break into chunks
    assert (  # TODO: this seems like an irrelevant copypasta since we use SearchQuery
        call_size <= 2000
    ), "Max number of calls ExportQuery can handle at a time is 2000."
    if len(ids) > call_size:
        chunked_ids = [ids[i : i + call_size] for i in range(0, len(ids), call_size)]
    else:
        chunked_ids = [
            ids,
        ]

    return chunked_ids


# File IO
def write_pickle(fn: str, data):
    with open(fn, "wb") as f:
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)


def read_pickle(fn: str):
    with open(fn, "rb") as f:
        data = pickle.load(f)
    return data


# various helper functions


def get_verbose(kwargs: dict):
    return kwargs["verbose"] if "verbose" in kwargs else False


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return str(msg) + "\n"
